/* beg_algo_entry

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% ipl_num_dil_geod
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\begin{algorithme}
{\bf Algorithme~: Erosion num\'erique g\'eod\'esique}
\medskip
\begin{verbatim}
int ipl_num_dil_geod (ipl_image *pt_in,
                      ipl_image **pt_out,
                      ipl_image *pt_in_ref,
                      int mode);
\end{verbatim}
\separ
Entr\'ee~:
	\begin{itemize}
	\item[] $I$, image num\'erique \`a dilater;
	\item[] $J$, image conditionnante;
	\end{itemize}
	\centerline{/ * La dilatation est calcul\'ee directement dans $I$ * /}
	\centerline{/ * En dehors de $I$, les pixels ont la valeur {\tt mode} * /}
Sortie~:
	\begin{itemize}
	\item[] $I$ image num\'erique dilat\'ee;
	\end{itemize}
\separ
Faire ind\'efiniment \{\\
\begin{itemize}
\item[] Parcourir $D_I$ en sens vid\'eo \{
	\begin{itemize}
	\item[] Soit $p$ le pixel courant;
	\item[] $I(p) \leftarrow \max\{I(p), I(p+\vec{u}_1),
			I(p+\vec{u}_2), I(p+\vec{u}_3) \}$;
	\item[] $I(p) \leftarrow \min \{ J(p), I(p) \};$
	\end{itemize}
\item[]	\}
\item[] Si $I$ n'a pas \'et\'e modifi\'ee, $exit$
\item[] Parcourir $D_I$ en sens vid\'eo inverse \{
	\begin{itemize}
	\item[] Soit $p$ le pixel courant;
	\item[] $I(p) \leftarrow \max\{I(p), I(p+\vec{u}_4),
			I(p+\vec{u}_5), I(p+\vec{u}_0) \}$;
	\item[] $I(p) \leftarrow \min \{ J(p), I(p) \};$
	\end{itemize} 
\item[]\}
\item[] Si $I$ n'a pas \'et\'e modifi\'ee, $exit$
\end{itemize}
\}\\
\end{algorithme}
   end_algo_entry */

#include "morpho.h"

int mm0_num_dil_geod (image_mark, image_ref, mode)
     IMAGE *image_mark, *image_ref;
     int   mode;
{
    /**********************************/
    /* les reconstructions numeriques */
    /**********************************/
    
    /*
      dilatation numerique, geodesique de mark dans ref
      mode est la valeur du cadre de mark avant le debut
      le resultat est dans mark (qui est donc modifiee)
      
      algo sequentiel
      */
     
    PIX *pt_mark, *pt_mark_old, *pt_ref, *pt_ref_old, pix_old;
    PIX *pt_vois2, *pt_vois6, maxim;
    int i, j, flag = 1, ret;
    int xmin, xmax, ymin, ymax;
    
    if ((ret = BASIC_CHECK (image_mark)) != RESULTAT_OK) return (ret);
    if ((ret = BASIC_CHECK (image_ref)) != RESULTAT_OK) return (ret);

    if ((ret = SET_SIZE (image_mark, image_ref))
	!= RESULTAT_OK) return (ret);
    
    xmin = image_mark->lx; xmax = xmin + image_mark->sx;
    ymin = image_mark->ly; ymax = ymin + image_mark->sy;
    
    if((ret=mm0_cadre (image_mark, mode))!=RESULTAT_OK) return (ret);
    
    while (1) {			/* forever */
	flag = 0;
	pt_mark = image_mark->bitmap  + xmin + image_mark->tx*ymin;
	pt_ref= image_ref->bitmap + xmin + image_ref->tx*ymin;
	for (j = ymin ; j < ymax; j++) {
	    
	    pt_mark_old = pt_mark;
	    pt_ref_old = pt_ref;
	    pt_vois2 = pt_mark - image_mark->tx + (j & 1);
	    
	    for (i = xmin ; i < xmax; i++, pt_mark++, pt_vois2++, pt_ref++) {
		pix_old = *pt_mark;
		maxim = (*pt_mark > *(pt_mark - 1)) ? *pt_mark : *(pt_mark - 1);
		if (*pt_vois2 > maxim) maxim = *pt_vois2;
		if (*(pt_vois2 - 1) > maxim) maxim = *(pt_vois2 - 1);
		*pt_mark = (maxim < *pt_ref) ? maxim : *pt_ref;
		if (pix_old != *pt_mark) flag = 1;
	    }
	    
	    pt_mark = pt_mark_old + image_mark->tx;
	    pt_ref = pt_ref_old + image_ref->tx;
	}
	
	if (flag == 0) return (RESULTAT_OK);
	
	flag = 0;
	pt_mark = image_mark->bitmap + (xmax - 1)+ image_mark->tx*(ymax - 1);
	pt_ref= image_ref->bitmap + (xmax - 1)+ image_ref->tx*(ymax - 1);
	for (j = ymax - 1 ; j >= ymin; j--) {
	    
	    pt_mark_old = pt_mark;
	    pt_ref_old = pt_ref;
	    pt_vois6 = pt_mark + image_mark->tx + (j & 1);
	    
	    for (i = xmax - 1 ; i >= xmin; i--, pt_mark--, pt_vois6--, pt_ref--) {
		pix_old = *pt_mark;
		maxim = (*pt_mark > *(pt_mark + 1)) ? *pt_mark : *(pt_mark + 1);
		if (*pt_vois6 > maxim) maxim = *pt_vois6;
		if (*(pt_vois6 - 1) > maxim) maxim = *(pt_vois6 - 1);
		*pt_mark = (maxim < *pt_ref) ? maxim : *pt_ref;
		if (pix_old != *pt_mark) flag = 1;
	    }
	    
	    pt_mark = pt_mark_old - image_mark->tx;
	    pt_ref = pt_ref_old - image_ref->tx;
	}
	
	if (flag == 0) return (RESULTAT_OK);
    }
}

int mm0_num_ero_geod (image_mark, image_ref, mode)
     IMAGE *image_mark, *image_ref; 
     int mode;
{
    /* le meme pour l'erosion */
    /* erode's successifs */

    PIX *pt_mark, *pt_mark_old, *pt_ref, *pt_ref_old, pix_old;
    PIX *pt_vois2, *pt_vois6, minim;
    int i, j, flag = 1, ret;
    int xmin, xmax, ymin, ymax;
    
    if ((ret = BASIC_CHECK (image_mark)) != RESULTAT_OK) {
#ifdef	LELISP
	(void) phimage(1,image_mark);
#endif
	return (ret);}
    if ((ret = BASIC_CHECK (image_ref)) != RESULTAT_OK) {
#ifdef	LELISP
	(void) phimage(2,image_ref);
#endif
	return (ret);}

    if ((ret = SET_SIZE (image_mark, image_ref))
	!= RESULTAT_OK) return (ret);
    
    xmin = image_mark->lx; xmax = xmin + image_mark->sx;
    ymin = image_mark->ly; ymax = ymin + image_mark->sy;
    
    if((ret=mm0_cadre (image_mark, mode))!=RESULTAT_OK) return (ret);
    
    while (1) {			/* forever */
	flag = 0;
	pt_mark = image_mark->bitmap  + xmin + image_mark->tx*ymin;
	pt_ref= image_ref->bitmap + xmin + image_ref->tx*ymin;
	for (j = ymin ; j < ymax; j++) {
	    
	    pt_mark_old = pt_mark;
	    pt_ref_old = pt_ref;
	    pt_vois2 = pt_mark - image_mark->tx + (j & 1);
	    
	    for (i = xmin ; i < xmax; i++, pt_mark++, pt_vois2++, pt_ref++) {
		pix_old = *pt_mark;
		minim = (*pt_mark < *(pt_mark - 1)) ? *pt_mark : *(pt_mark - 1);
		if (*pt_vois2 < minim) minim = *pt_vois2;
		if (*(pt_vois2 - 1) < minim) minim = *(pt_vois2 - 1);
		*pt_mark = (minim > *pt_ref) ? minim : *pt_ref;
		if (pix_old != *pt_mark) flag = 1;
	    }
	    
	    pt_mark = pt_mark_old + image_mark->tx;
	    pt_ref = pt_ref_old + image_ref->tx;
	}
	
	if (flag == 0) return (RESULTAT_OK);
	
	flag = 0;
	pt_mark = image_mark->bitmap + (xmax - 1)+ image_mark->tx*(ymax - 1);
	pt_ref= image_ref->bitmap + (xmax - 1)+ image_ref->tx*(ymax - 1);
	for (j = ymax - 1 ; j >= ymin; j--) {
	    
	    pt_mark_old = pt_mark;
	    pt_ref_old = pt_ref;
	    pt_vois6 = pt_mark + image_mark->tx + (j & 1);
	    
	    for (i = xmax - 1 ; i >= xmin; i--, pt_mark--, pt_vois6--, pt_ref--) {
		pix_old = *pt_mark;
		minim = (*pt_mark < *(pt_mark + 1)) ? *pt_mark : *(pt_mark + 1);
		if (*pt_vois6 < minim) minim = *pt_vois6;
		if (*(pt_vois6 - 1) < minim) minim = *(pt_vois6 - 1);
		*pt_mark = (minim > *pt_ref) ? minim : *pt_ref;
		if (pix_old != *pt_mark) flag = 1;
	    }
	    
	    pt_mark = pt_mark_old - image_mark->tx;
	    pt_ref = pt_ref_old - image_ref->tx;
	}
	
	if (flag == 0) return (RESULTAT_OK);
    }
}
