/* beg_algo_entry

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% ipl_dilate
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\begin{algorithme}
{\bf Algorithme~: Dilatation num\'erique par un hexagone}
\medskip
\begin{verbatim}
int ipl_dilate (ipl_image *pt_in,
                ipl_image **pt_out,
                int size,
                int mode);
\end{verbatim}
\separ
Entr\'ee~:
	\begin{itemize}
	\item[] $I$, image num\'erique;
	\end{itemize}
\centerline{/ * {\tt mode} doit \^etre $0$ * /}
Sortie~:
	\begin{itemize}
	\item[] $J$, image num\'erique dilat\'ee de taille {\tt size};
	\end{itemize}
\separ
Faire {\tt size} fois \{
\begin{itemize} 
\item[] Parcourir $D_I$ en sens vid\'eo \{
	\begin{itemize}
	\item[] Soit $p$ le pixel courant;
	\item[] Si $J(p) \leftarrow \max\left\{
			\begin{array}{l}
			I(p), I(p+\vec{u}_0), I(p+\vec{u}_1), I(p+\vec{u}_2),\\
			I(p+\vec{u}_3),	I(p+\vec{u}_4), I(p+\vec{u}_5) 
			\end{array}\right\}$;
	\end{itemize}
\item[] \}
\item[] Copier $J$ dans $I$;
\end{itemize}
\centerline{/ * ou mieux, \'echanger les noms $I$ et $J$ * /}
\}\\
\end{algorithme}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
end_algo_entry */

#include "morpho.h"

/* ---------------- basic routine: dilation by one ---------------- */

/* NE PAS APPELER DIRECTEMENT CETTE ROUTINE */
/* "in" et "out" doivent etre differentes */
/* le cadre de "in" est modifie */

int mm0_dilate1 (in, out)
     IMAGE *in, *out;
{
    PIX *pt_in, *pt_out, *pt_end;
    PIX *pt_vois2, *pt_vois6, pred, maxim;
    int j;

    if (in == out) return (IMAGE_NOT_DIFFERENT);
    mm0_cadre (in, PIX_MIN);

    for (j = in->ly ; j < in->ly + in->sy ; j++) {
	pt_in = in->bitmap + in->lx + j*in->tx;
	pt_end = pt_in + in->sx;
	pt_out = out->bitmap + out->lx + j*out->tx;

	pt_vois2 = pt_in - in->tx + (j & 1);
	pt_vois6 = pt_in + in->tx + (j & 1);
	
	if (j & 1) {
	    pred = (*(pt_vois2-1) > *(pt_vois6-1))
		? *(pt_vois2-1) : *(pt_vois6-1);
	} else pred = *pt_in;

	for (; pt_in < pt_end; pt_in++, pt_vois2++, pt_vois6++) {
	    maxim = (*pt_in > *(pt_in + 1)) ? *pt_in : *(pt_in + 1);
	    if (*pt_vois2 > maxim) maxim = *pt_vois2;
	    if (*pt_vois6 > maxim) maxim = *pt_vois6;
	    *pt_out++ = (maxim > pred) ? maxim : pred;
	    pred = maxim;
	}
    }
    return (RESULTAT_OK);
}

int mm0_dilate (in, out, work, taille)
     IMAGE *in, *out, *work;
     int taille;
{
    /* "in", "out", "work" doivent etre differentes */
    /* le cadre de "in" est modifie */

    IMAGE *im_tmp_1, *im_tmp_2, *im_tmp;
    int i, ret;

    if ((ret = BASIC_CHECK (in))   != RESULTAT_OK) return (ret);
    if ((ret = SET_SIZE (in,out))  != RESULTAT_OK) return (ret);
    if ((ret = SET_SIZE (in,work)) != RESULTAT_OK) return (ret);
    /* allocation could be wrong: check further */
    if ((ret = BASIC_CHECK (out))  != RESULTAT_OK) return (ret);
    if ((ret = BASIC_CHECK (work)) != RESULTAT_OK) return (ret);

    if (taille <= 0) {
	return( mm0_copy (in, out));
    } else {
	/* operate between work & out image, for odd size dilations
	   one operates from work to out */
	if ((taille & 1) == 0) {
	    im_tmp_2 = work; im_tmp_1 = out;
	} else {
	    im_tmp_2 = out; im_tmp_1 = work;
	}
	if((ret=mm0_copy (in, im_tmp_1))!=RESULTAT_OK) return(ret);

	for (i = 0; i < taille; i++) {
	    if ((ret = mm0_dilate1 (im_tmp_1, im_tmp_2)) != RESULTAT_OK)
	      return (ret);
	    /* exchange input/output */
	    im_tmp = im_tmp_1; im_tmp_1 = im_tmp_2; im_tmp_2 = im_tmp;
	}
    }
    return (RESULTAT_OK);
}

/* erosion */

/* beg_algo_entry

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% ipl_erode
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\begin{algorithme}
{\bf Algorithme~: Erosion num\'erique par un hexagone}
\medskip
\begin{verbatim}
int ipl_erode (ipl_image *pt_in,
                ipl_image **pt_out,
                int size,
                int mode);
\end{verbatim}
\separ
Entr\'ee~:
	\begin{itemize}
	\item[] $I$, image num\'erique;
	\end{itemize}
\centerline{/ * {\tt mode} doit \^etre $0$ * /}
Sortie~:
	\begin{itemize}
	\item[] $J$, image num\'erique 'erod\'ee de taille {\tt size};
	\end{itemize}
\separ
Faire {\tt size} fois \{
\begin{itemize} 
\item[] Parcourir $D_I$ en sens vid\'eo \{
	\begin{itemize}
	\item[] Soit $p$ le pixel courant;
	\item[] Si $J(p) \leftarrow \min\left\{
			\begin{array}{l}
			I(p), I(p+\vec{u}_0), I(p+\vec{u}_1), I(p+\vec{u}_2),\\
			I(p+\vec{u}_3),	I(p+\vec{u}_4), I(p+\vec{u}_5) 
			\end{array}\right\}$;
	\end{itemize}
\item[] \}
\item[] Copier $J$ dans $I$;
\end{itemize}
\centerline{/ * ou mieux, \'echanger les noms $I$ et $J$ * /}
\}\\
\end{algorithme}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   end_algo_entry */

/* NE PAS APPELER DIRECTEMENT CETTE ROUTINE */
/* "in" et "out" doivent etre differentes */
/* le cadre de "in" est modifie */

int mm0_erode1 (in, out)
	IMAGE *in, *out;
{
    PIX *pt_in, *pt_out, *pt_end;
    PIX *pt_vois2, *pt_vois6, pred, minim;
    int j;
    
    if (in == out) return (IMAGE_NOT_DIFFERENT);
    (void)mm0_cadre (in, PIX_MAX);
    for (j = in->ly; j < in->ly + in->sy; j++) {
	pt_in = in->bitmap + in->lx + j*in->tx;
	pt_end = pt_in + in->sx;
	pt_out = out->bitmap + out->lx + j*out->tx;
	pt_vois2 = pt_in - in->tx + (j & 1);
	pt_vois6 = pt_in + in->tx + (j & 1);
	
	if (j & 1) {
	    pred = (*(pt_vois2-1) < *(pt_vois6-1))
	      ? *(pt_vois2-1) : *(pt_vois6-1);
	} else pred = *pt_in;
	
	for (; pt_in < pt_end; pt_in++, pt_vois2++, pt_vois6++) {
	    minim = (*pt_in < *(pt_in + 1)) ? *pt_in : *(pt_in + 1);
	    if (*pt_vois2 < minim) minim = *pt_vois2;
	    if (*pt_vois6 < minim) minim = *pt_vois6;
	    *pt_out++ = (minim < pred) ? minim : pred;
	    pred = minim;
	}
    }
    
    return (RESULTAT_OK);
}

int mm0_erode (in, out, work, taille)
     IMAGE *in, *out, *work;
     int taille;
{

    /* "in", "out", "work" doivent etre differentes */
    /* le cadre de "in" est modifie */
    
    IMAGE *im_tmp_1, *im_tmp_2, *im_tmp;
    int i, ret;
    
    
    if ((ret = BASIC_CHECK (in))   != RESULTAT_OK) return (ret);
    if ((ret = SET_SIZE (in,out))  != RESULTAT_OK) return (ret);
    if ((ret = SET_SIZE (in,work)) != RESULTAT_OK) return (ret);
    /* allocation could be wrong: check further */
    if ((ret = BASIC_CHECK (out))  != RESULTAT_OK) return (ret);
    if ((ret = BASIC_CHECK (work)) != RESULTAT_OK) return (ret);
    
    if (taille <= 0) {
	return( mm0_copy (in, out));
    } else {
	/* operate between work & out image, for odd size dilations
	   one operates from work to out */
	if ((taille & 1) == 0) {
	    im_tmp_2 = work; im_tmp_1 = out;
	} else {
	    im_tmp_2 = out; im_tmp_1 = work;
	}
	if((ret=mm0_copy (in, im_tmp_1))!=RESULTAT_OK) return(ret);
    
	for (i = 0; i < taille; i++) {
	    if ((ret = mm0_erode1 (im_tmp_1, im_tmp_2)) != 0) return (ret);
	    /* exchange input/output */
	    im_tmp = im_tmp_1; im_tmp_1 = im_tmp_2; im_tmp_2 = im_tmp;
	}
    }    
    return (RESULTAT_OK);
}

/* beg_algo_entry

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% ipl_open
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\begin{algorithme}
{\bf Algorithme~: Ouverture num\'erique par un hexagone}
\medskip
\begin{verbatim}
int ipl_open (ipl_image *pt_in,
                ipl_image **pt_out,
                int size,
                int mode);
\end{verbatim}
\separ
Entr\'ee~:
	\begin{itemize}
	\item[] $I$, image num\'erique;
	\end{itemize}
\centerline{/ * {\tt mode} doit \^etre $0$ * /}
Sortie~:
	\begin{itemize}
	\item[] $J$, image num\'erique ouverte de taille {\tt size};
	\end{itemize}
\separ
Eroder $I$ dans $J$ de taille {\tt size}\\
Dilater $J$ dans $J$ de taille {\tt size}\\
\end{algorithme}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   end_algo_entry */

int	 mm0_close ( pt_in, pt_out, work, size)
     IMAGE *pt_in;
     IMAGE *pt_out;
     IMAGE *work;
     int   size;
{
    int res;
    
    /* do the dilatations */
    res = mm0_dilate( pt_in, pt_out, work, size);
    if(res != RESULTAT_OK) return(res);
    /* do the erosions */
    res = mm0_erode ( pt_out, pt_out, work, size);
    return ( res );
}

int	 mm0_open ( pt_in, pt_out, work, size)
     IMAGE *pt_in;
     IMAGE *pt_out;
     IMAGE *work;
     int   size;
{
    int res;
    
    res = mm0_erode( pt_in, pt_out, work, size);
    if(res != RESULTAT_OK) return(res);
    res = mm0_dilate ( pt_out, pt_out, work, size);
    return ( res );
}

