#include "morpho.h"
/* les routines directes */
/*************************/

int mm0_add (in1, in2, out)
IMAGE *in1, *in2, *out;
{
  return (mm0_algebre (in1, in2, out, CG_ADD));
}

int mm0_sub (in1, in2, out)
IMAGE *in1, *in2, *out;
{
  return (mm0_algebre (in1, in2, out, CG_SUB));
}

int mm0_mul (in1, in2, out)
IMAGE *in1, *in2, *out;
{
  return (mm0_algebre (in1, in2, out, CG_MUL));
}

int mm0_div (in1, in2, out)
IMAGE *in1, *in2, *out;
{
  return (mm0_algebre (in1, in2, out, CG_DIV));
}

int mm0_and (in1, in2, out)
IMAGE *in1, *in2, *out;
{
  return (mm0_algebre (in1, in2, out, CG_AND));
}

int mm0_or (in1, in2, out)
IMAGE *in1, *in2, *out;
{
  return (mm0_algebre (in1, in2, out, CG_OR));
}

int mm0_xor (in1, in2, out)
IMAGE *in1, *in2, *out;
{
  return (mm0_algebre (in1, in2, out, CG_XOR));
}

int mm0_andnot (in1, in2, out)
IMAGE *in1, *in2, *out;
{
  return (mm0_algebre (in1, in2, out, CG_ANDNOT));
}

int mm0_max (in1, in2, out)
IMAGE *in1, *in2, *out;
{
  return (mm0_algebre (in1, in2, out, CG_MAX));
}

int mm0_min (in1, in2, out)
IMAGE *in1, *in2, *out;
{
  return (mm0_algebre (in1, in2, out, CG_MIN));
}


/* ---------------------------------------------------------------- */
int mm0_algebre (in1, in2, out, type)
     IMAGE *in1, *in2, *out;
     int type;
     
     /* op. algebriques entre in1 et in2 */
     
{
    PIX *pt_in1, *pt_in1_old, *pt_in2, *pt_in2_old, *pt_out, *pt_out_old;
    int i, j, ret;
    
    if ((ret = BASIC_CHECK (in1)) != RESULTAT_OK) return (ret);
    if ((ret = BASIC_CHECK (in2)) != RESULTAT_OK) return (ret);
    if ((ret = BASIC_CHECK (out)) != RESULTAT_OK)  return (ret);
    if ((ret = SIZE_CHECK (in1,in2)) != RESULTAT_OK) return (ret);
    if ((ret = SET_SIZE (in1,out)) != RESULTAT_OK) return (ret);

    pt_in1 = BEG_PIXPTR(in1);
    pt_in2 = BEG_PIXPTR(in2);
    pt_out = BEG_PIXPTR(out);
    
    for (j = 0; j < in1->sy; j++) {
	
	pt_in1_old = pt_in1;
	pt_in2_old = pt_in2;
	pt_out_old = pt_out;
	switch (type)
	  {
	      default:		/* argument not in allowed range */
	      return (PARAM_NOT_IN_RANGE);
	      
	      case CG_ADD : 
		for (i = 0 ; i < in1->sx; i++, pt_in1++, pt_in2++, pt_out++)
		  *pt_out = *pt_in1 + *pt_in2;
	      break;
	      case CG_SUB : 
		for (i = 0 ; i < in1->sx; i++, pt_in1++, pt_in2++, pt_out++)
		  *pt_out = *pt_in1 - *pt_in2;
	      break;
	      case CG_MUL : 
		for (i = 0 ; i < in1->sx; i++, pt_in1++, pt_in2++, pt_out++)
		  *pt_out = *pt_in1 * *pt_in2;
	      break;
	      case CG_DIV : 
		for (i = 0 ; i < in1->sx; i++, pt_in1++, pt_in2++, pt_out++)
		  *pt_out = *pt_in1 / *pt_in2;
	      break;
	      case CG_AND : 
		for (i = 0 ; i < in1->sx; i++, pt_in1++, pt_in2++, pt_out++)
		  *pt_out = *pt_in1 & *pt_in2;
	      break;
	      case CG_OR : 
		for (i = 0 ; i < in1->sx; i++, pt_in1++, pt_in2++, pt_out++)
		  *pt_out = *pt_in1 | *pt_in2;
	      break;
	      case CG_XOR : 
		for (i = 0 ; i < in1->sx; i++, pt_in1++, pt_in2++, pt_out++)
		  *pt_out = *pt_in1 ^ *pt_in2;
	      break;
	      case CG_ANDNOT : 
		for (i = 0 ; i < in1->sx; i++, pt_in1++, pt_in2++, pt_out++)
		  *pt_out = *pt_in1 & (1 -  *pt_in2);
	      break;
	      case CG_MAX :
		for (i = 0 ; i < in1->sx; i++, pt_in1++, pt_in2++, pt_out++)
		  *pt_out = (*pt_in1 > *pt_in2) ? *pt_in1 : *pt_in2;
	      break;
	      case CG_MIN :
		for (i = 0 ; i < in1->sx; i++, pt_in1++, pt_in2++, pt_out++)
		  *pt_out = (*pt_in1 < *pt_in2) ? *pt_in1 : *pt_in2;
	      break;
	  }
	pt_in1 = pt_in1_old + in1->tx;
	pt_in2 = pt_in2_old + in2->tx;
	pt_out = pt_out_old + out->tx;
    }
    
    return (RESULTAT_OK);
}
