#include "morpho.h"

int mm0_cadd (in1, cte, out)
IMAGE *in1, *out;
int cte;
{
  return (mm0_calgebre (in1, cte, out, CG_ADD));
}

int mm0_csub (in1, cte, out)
IMAGE *in1, *out;
int cte;
{
  return (mm0_calgebre (in1, cte, out, CG_SUB));
}

int mm0_cmul (in1, cte, out)
IMAGE *in1, *out;
int cte;
{
  return (mm0_calgebre (in1, cte, out, CG_MUL));
}

int mm0_cdiv (in1, cte, out)
IMAGE *in1, *out;
int cte;
{
  return (mm0_calgebre (in1, cte, out, CG_DIV));
}

int mm0_cand (in1, cte, out)
IMAGE *in1, *out;
int cte;
{
  return (mm0_calgebre (in1, cte, out, CG_AND));
}

int mm0_cor (in1, cte, out)
IMAGE *in1, *out;
int cte;
{
  return (mm0_calgebre (in1, cte, out, CG_OR));
}

int mm0_cxor (in1, cte, out)
IMAGE *in1, *out;
int cte;
{
  return (mm0_calgebre (in1, cte, out, CG_XOR));
}

int mm0_candnot (in1, cte, out)
IMAGE *in1, *out;
int cte;
{
  return (mm0_calgebre (in1, cte, out, CG_ANDNOT));
}

int mm0_cmax (in1, cte, out)
IMAGE *in1, *out;
int cte;
{
  return (mm0_calgebre (in1, cte, out, CG_MAX));
}

int mm0_cmin (in1, cte, out)
IMAGE *in1, *out;
int cte;
{
  return (mm0_calgebre (in1, cte, out, CG_MIN));
}

/* ---------------------------------------------------------------- */
/* op. algebriques in2 est une constante */
int mm0_calgebre (in1, int_in2, out, type)
     IMAGE *in1, *out;
     int type, int_in2;
{
    PIX *pt_in1, *pt_in1_old, *pt_out, *pt_out_old;
    PIX pix_in2=int_in2;
    int i, j, ret;
    
    if ((ret = BASIC_CHECK (in1)) != RESULTAT_OK) return (ret);
    if ((ret = BASIC_CHECK (out)) != RESULTAT_OK) return (ret);
    if ((ret = SET_SIZE (in1,out)) != RESULTAT_OK) return (ret);
    
    pt_in1 = BEG_PIXPTR(in1);
    pt_out = BEG_PIXPTR(out);
    
    for (j = 0; j < in1->sy; j++) {
	pt_in1_old = pt_in1;
	pt_out_old = pt_out;
	
	switch (type)
	  {
	      default:		/* argument not in allowed range */
	      return(CG_PARAM_NOTIR);
	      
	      case CG_ADD : 
		for (i = 0 ; i < in1->sx; i++, pt_in1++, pt_out++)
		  *pt_out = *pt_in1 + pix_in2;
	      break;
	      case CG_SUB : 
		for (i = 0 ; i < in1->sx; i++, pt_in1++, pt_out++)
		  *pt_out = *pt_in1 - pix_in2;
	      break;
	      case CG_MUL : 
		for (i = 0 ; i < in1->sx; i++, pt_in1++, pt_out++)
		  *pt_out = *pt_in1 * pix_in2;
	      break;
	      case CG_DIV : 
		for (i = 0 ; i < in1->sx; i++, pt_in1++, pt_out++)
		  *pt_out = *pt_in1 / pix_in2;
	      break;
	      case CG_AND : 
		for (i = 0 ; i < in1->sx; i++, pt_in1++, pt_out++)
		  *pt_out = *pt_in1 & pix_in2;
	      break;
	      case CG_OR : 
		for (i = 0 ; i < in1->sx; i++, pt_in1++, pt_out++)
		  *pt_out = *pt_in1 | pix_in2;
	      break;
	      case CG_XOR : 
		for (i = 0 ; i < in1->sx; i++, pt_in1++, pt_out++)
		  *pt_out = *pt_in1 ^ pix_in2;
	      break;
	      case CG_ANDNOT : 
		for (i = 0 ; i < in1->sx; i++, pt_in1++, pt_out++)
		  *pt_out = *pt_in1 & (1 -  pix_in2);
	      break;
	      case CG_MAX : 
		for (i = 0 ; i < in1->sx; i++, pt_in1++, pt_out++)
		  *pt_out = (*pt_in1 > pix_in2) ? *pt_in1 : pix_in2;
	      break;
	      case CG_MIN :
		for (i = 0 ; i < in1->sx; i++, pt_in1++, pt_out++)
		  *pt_out = (*pt_in1 < pix_in2) ? *pt_in1 : pix_in2;
	      break;
	  }
	pt_in1 = pt_in1_old + in1->tx;
	pt_out = pt_out_old + out->tx;
    }
    return (RESULTAT_OK);
}
