/* beg_algo_entry

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% ipl_sucero
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\begin{algorithme}
{\bf Algorithme~: Erodes successifs}
\medskip
\begin{verbatim}
int ipl_sucero (ipl_image *pt_in,
                ipl_image **pt_out,
                int mode );
\end{verbatim}
\separ
Entr\'ee~:
	\begin{itemize}
	\item[] $I$, image binaire;
	\end{itemize}
\centerline{/ * La fonction distance est calcul\'ee directement
		dans, $I$ * /}
\centerline{/ * En dehors de $I$, les pixels ont la valeur $0$ si ${\tt
mode}=1$ et $1$ sinon * /}
Sortie~:
	\begin{itemize}
	\item[] $I$, image num\'erique des \'erod\'es successifs;
	\end{itemize}
\separ
Parcourir $D_I$ en sens vid\'eo \{
\begin{itemize}
\item[] Soit $p$ le pixel courant;
\item[] Si $I(p) \neq 0$ alors
	\begin{itemize}
	\item[] $I(p) \leftarrow \inf\{I(p+\vec{u}_1) + 1,
		I(p+\vec{u}_2) + 1, I(p+\vec{u}_3) + 1 \}$;
	\end{itemize}
\end{itemize}
\}\\
Parcourir $D_I$ en sens vid\'eo inverse \{
\begin{itemize}
\item[] Soit $p$ le pixel courant;
\item[] Si $I(p) \neq 0$ alors
	\begin{itemize}
	\item[] $I(p) \leftarrow \inf\{I(p), I(p+\vec{u}_4) + 1,
		I(p+\vec{u}_5) + 1, I(p+\vec{u}_0) + 1 \}$;
	\end{itemize}
\end{itemize}
\}\\
\end{algorithme}
   end_algo_entry */

#include "morpho.h"

/* erode's successifs */
int mm0_sucero (image_in, image_out, mode)
     int   mode;
     IMAGE *image_in, *image_out;
{
    PIX *pt_in, *pt_in_old, *pt_out, *pt_out_old;
    PIX *pt_vois2, *pt_vois6, minim;

    int i, j, ret;
    int xmin, xmax, ymin, ymax, lincr;

    if ((ret = BASIC_CHECK (image_in )) != RESULTAT_OK) return (ret);
    if ((ret = BASIC_CHECK (image_out)) != RESULTAT_OK) return (ret);
    if ((ret = ALLOC_X_CHECK (image_in,image_out)) != RESULTAT_OK) return (ret);
    if ((ret = SET_SIZE (image_in,image_out)) != RESULTAT_OK) return (ret);
    lincr = image_in->tx;

    xmin = image_in->lx; xmax = xmin + image_in->sx;
    ymin = image_in->ly; ymax = ymin + image_in->sy;
    pt_in = image_in->bitmap  + xmin + lincr*ymin;
    pt_out= image_out->bitmap + xmin + lincr*ymin;

    (void)mm0_cadre (image_out, (mode == 0) ? PIX_MAX : 0);

    /* passe en sens video */

    for (j = ymin ; j < ymax; j++) {
	pt_in_old = pt_in;
	pt_out_old = pt_out;
	pt_vois2 = pt_out - lincr + (j & 1);
	
	for (i = xmin ; i < xmax; i++, pt_in++, pt_vois2++, pt_out++) {
	    if (*pt_in == 0) *pt_out = 0;
	    else {
		minim = (*(pt_vois2 - 1) < *pt_vois2) ? *(pt_vois2 - 1) : *pt_vois2;
		if (*(pt_out - 1) < minim) minim = *(pt_out - 1);
		if (minim != PIX_MAX) minim++;
		*pt_out = minim;
	    }
	}
	pt_in = pt_in_old + lincr;
	pt_out = pt_out_old + lincr;
    }
    
    /* passe en sens video inverse */
    
    pt_in = image_out->bitmap + (xmax - 1)+ lincr*(ymax - 1);
    pt_out= image_out->bitmap + (xmax - 1)+ lincr*(ymax - 1);
    
    for (j = ymax - 1 ; j >= ymin; j--) {
	pt_in_old = pt_in;
	pt_out_old = pt_out;
	pt_vois6 = pt_out + lincr + (j & 1);

	for (i = xmax - 1 ; i >= xmin; i--, pt_in--, pt_vois6--, pt_out--) {
	    if (*pt_in != 0) {
		minim = (*(pt_vois6 - 1) < *pt_vois6) ? *(pt_vois6 - 1) : *pt_vois6;
		if (*(pt_out + 1) < minim) minim = *(pt_out + 1);
		if (minim != PIX_MAX) minim++;
		if (*pt_out > minim) *pt_out = minim;
	    }
	}
	pt_in = pt_in_old - lincr;
	pt_out = pt_out_old - lincr;
    }
    return (RESULTAT_OK);
}

int mm0_distance_hexa (inout)
     IMAGE *inout;
{
    int dist = 1;
    int ret;
    GLACE  *cgl, *mm0_alloc_chains();

    cgl=mm0_alloc_chains (inout);
    if(cgl==NULL) return(LACET_ALLOC_FAILED);

    ret=mm0_cont (inout, cgl);
    if(ret!=RESULTAT_OK) return(ret);

    (void)mm0_cadre (inout, 1);

    cgl->imavide = 0;

    while (cgl->imavide == 0) {
	mm0_dil_chaine (cgl);
	mm0_adjust_masque (inout, cgl, ++dist, 0);
    }

    return (RESULTAT_OK);
}

mm0_distance_dodeca (inout)
     IMAGE *inout;
{
    
    int dist = 1, i1, i2;
    int ret;
    GLACE  *cgl, *mm0_alloc_chains();

    cgl=mm0_alloc_chains (inout);
    if(cgl==NULL) return(LACET_ALLOC_FAILED);

    ret=mm0_cont (inout, cgl);
    if(ret!=RESULTAT_OK) return(ret);

    (void)mm0_cadre (inout, 1);
    (void)mm0_cadre_int (inout, 1);
    cgl->imavide = 0;
    i1 = i2 = 0;
    
    while (cgl->imavide == 0) {
	if (i1 + 1 < 1.732 * i2) {
	    i1 += 1;
	    mm0_dil_chaine (cgl);
	    mm0_adjust_masque (inout, cgl, ++dist, 0);
	} else {
	    i2 += 1;
	    mm0_dil_chaine (cgl);
	    mm0_adjust_masque (inout, cgl, ++dist, 0);
	    mm0_dil_chaine_conj (cgl);
	    mm0_trace_masque (inout, cgl, dist, 2, 0);
	    mm0_copy_chaine (cgl, 2, 1);
	}
    }
    
    return (RESULTAT_OK);
}

/* ======================================================================
\operation{(*distance\_geod inout masque)}
{calcule la distance g\'eod\'esique \`a l'image binaire {\tt inout} en
utilisant la distance g\'eod\'esique d\'efinie par l'image {\tt
masque}.}
====================================================================== */
int mm0_distance_geod (inout, mask)
     IMAGE *inout, *mask;
{

    int dist = 3;
    int ret;
    GLACE *cgl, *mm0_alloc_chains();

    cgl=mm0_alloc_chains (inout);
    if(cgl==NULL) return(LACET_ALLOC_FAILED);

    ret=mm0_cont (inout, cgl);
    if(ret!=RESULTAT_OK) return(ret);

    ret=mm0_algebre (inout, mask, inout, CG_ADD);
    if(ret!=RESULTAT_OK) return(ret);

    (void)mm0_cadre (inout, 0);

    cgl->imavide = 0;
    while (cgl->imavide == 0) {
	mm0_dil_chaine (cgl);
	mm0_adjust_masque (inout, cgl, ++dist, 1);
    }

    return (RESULTAT_OK);
}
