/* beg_algo_entry

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% ipl_hmtbin
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\begin{algorithme}
{\bf Algorithme~: Transformation g\'en\'erale sur un voisinage de taille $1$}
\medskip
\begin{verbatim}
int ipl_hmtbin (ipl_image *pt_in,
                ipl_image **pt_out,
                int el1, el2, el3, el4, el5, el6, el7,
                int taille,
                int rot,
                int mode);
\end{verbatim}
\separ
Entr\'ee~:
	\begin{itemize}
	\item[] $I$, image binaire;
	\item[] {\tt el}, \'el\'ement structurant de taille $1$;

	\centerline{/ * {\tt el1} centre, {\tt el2} $\rightarrow$ {\tt
		el7} sens trigonom\'ertrique * /}
	\item[] {\tt taille}, nombre de balayage d'image;
	\item[] {\tt rot} = 0, \'el\'ement structurant fixe;
	\item[] {\tt ~~~} = 1, \'el\'ement structurant tournant de
		$60$ degr\'es \`a chaque balayage;
	\item[] {\tt mode} = 3, tout ou rien;
	\item[] {\tt ~~~~} = 4, amincissement;
	\item[] {\tt ~~~~} = 5, \'epaississement;
	\end{itemize}
Sortie~:
	\begin{itemize}
	\item[] $J$, image binaire r\'esultat;
	\end{itemize}
\separ
Faire {\tt taille} fois \{
\begin{itemize} 
\item[] Parcourir $D_I$ en sens vid\'eo \{
	\begin{itemize}
	\item[] Soit $p$ le pixel courant, $N_I(p)$ son voisinage dans $I$;
	\item[] Selon {\tt mode} \{
		\begin{itemize}
		\item[] \makebox[2.6cm][l]{tout ou rien~:}
			Si ${\tt el} = N_I(p)$, $J(p) \leftarrow 1$,
			sinon $J(p) \leftarrow 0$;
		\item[] \makebox[2.6cm][l]{amincissement~:}
			Si ${\tt el} = N_I(p)$, $J(p) \leftarrow 0$,
			sinon $J(p) \leftarrow I(p)$;
		\item[] \makebox[2.6cm][l]{\'epaississement~:}
			Si ${\tt el} = N_I(p)$, $J(p) \leftarrow 1$,
			sinon $J(p) \leftarrow I(p)$;
		\end{itemize}
	\item[] \}
	\end{itemize}
\item[] \}
\item[] Copier $J$ dans $I$; / * ou mieux, \'echanger les noms $I$ et $J$ * /
\end{itemize}
\}
\end{algorithme}
   end_algo_entry */

#include "morpho.h"

/* ================================================================ */

int mm0_hmtstruct (im_in, im_out, elt_struct, mode, modif)
    IMAGE *im_in, *im_out;
    int *elt_struct, mode, *modif;
{
    PIX lut[128];

    (void) cg_fillut (elt_struct, lut, mode);
    return ( mm0_hmtlut (im_in, im_out, lut, mode, modif));
}

int mm0_hmtbin (im_in, im_out, elt_struct, taille, rot, mode)
    IMAGE *im_in, *im_out;
    int   *elt_struct, taille, rot, mode;
{
/***************************************************/
/*                                                 */
/*        TRANSFORMATION BINAIRE GENERALE          */
/*                                                 */
/*   in         : image d'entree                   */
/*   out        : image de sortie                  */
/*   elt_struct : vecteur de l'elt structurant     */
/*   taille     : nombre d'iterations              */
/*   rot  = 0   : rotation de elt_struct           */
/*        = 1   : pas de rotation                  */
/*   mode = 1   : erosion        #[1 1 2 2 2 2 2]  */
/*        = 2   : dilatation     #[1 1 2 2 2 2 2]  */
/*        = 3   : ring           #[1 2 1 1 1 2 0]  */
/*        = 4   : amincissement  #[1 2 1 1 2 0 0]  */
/*        = 5   : epaississement #[0 2 1 1 2 0 0]  */
/*                                                 */
/***************************************************/

    int modifie, modifie_un, i, j, bord, xxx, res;

    if (taille < 0) {
	if (im_out != im_in) {
	    return ( mm0_copy (im_in, im_out));
	} else {
	    return (RESULTAT_OK);
	}
    }

    if (taille == 0) taille = im_in->tx * 100; /* until converging */
    for (i = 0; i < taille; i++) {
	modifie = 0;
	if (rot != 0) {
	    if (i != 0) im_in = im_out;
	    res = mm0_hmtstruct (im_in, im_out, elt_struct, mode, &modifie);
	    if(res != RESULTAT_OK) return(res);
	    /* fprintf(stderr,"Nombre de points modifie's %d\n\r", modifie); */
	    if (modifie == 0) return (RESULTAT_OK);
	} else {
	    for (j = 0; j < 6; j++) {
		if (j != 0) im_in = im_out;
		modifie_un = 0;
		res = mm0_hmtstruct (im_in, im_out, elt_struct, mode, &modifie_un);
		if(res != RESULTAT_OK) return(res);
		/* fprintf(stderr,"Nombre de points modifie's %d\n\r", modifie); */
		if (modifie_un != 0) modifie = 1;
		(void) cg_rotation (elt_struct);
	    }
	    if (modifie == 0) return (RESULTAT_OK);
	}
    }

    return (RESULTAT_OK);
}

int mm0_hmtlut (in, out, lut, mode, modif)
     IMAGE  *in,  *out;
     int    mode, *modif;
     PIX    *lut;
{
    PIX     **buf_lines = 0;
    PIX     *pt_in, *pt_out, *pt_vois2, *pt_vois6, centre=0;
    int     i, j, bord=1, xxx, ret, temp;
    int     haut = 0, milieu = 1, bas = 2;
    int     xmin, xmax, ymin, ymax;


    if ((ret = BASIC_CHECK (in)) != RESULTAT_OK) return (ret);
    if ((ret = BASIC_CHECK (out)) != RESULTAT_OK) return (ret);
    
    if ((ret = SET_SIZE (in,out)) != RESULTAT_OK) return (ret);

    buf_lines=(PIX **)mm0_alloc_lbuf(in);
    if(buf_lines==NULL) return(ALLOC_FAILED);

    xmin = in->lx; xmax = xmin + in->sx;
    ymin = in->ly; ymax = ymin + in->sy;

    switch (mode) {
    case 1 : bord = 1; centre = 0; break;
    case 2 : bord = 0; centre = 1; break;
    case 3 : bord = 0; centre = 2; break;
    case 4 : bord = 1; centre = 0; break;
    case 5 : bord = 0; centre = 1; break;
    default:  return(CG_PARAM_NOTIR);
    }
    ret = mm0_cadre (in, bord);
    if(ret != RESULTAT_OK) return(ret);

    *modif = 0;

    (void) mm0_readim (in, ymin - 1, buf_lines[haut]);
    (void) mm0_readim (in, ymin, buf_lines[milieu]);

    for (j = ymin ; j < ymax; j++) {

	(void) mm0_readim (in, j + 1, buf_lines[bas]);
	pt_in = buf_lines[milieu] + xmin;
	pt_out = buf_lines[3] + xmin;
	pt_vois2 = buf_lines[haut] + xmin + (j & 1);
	pt_vois6 = buf_lines[bas] + xmin + (j & 1);

	for (i = xmin ; i < xmax; i++) {
	    if (*pt_in == centre) *pt_out = centre;
	    else {
		xxx = *pt_in | *(pt_in + 1) << 1 | *(pt_in - 1) << 4
		    | *pt_vois2 << 2 | *(pt_vois2 - 1) << 3
		    | *pt_vois6 << 6 | *(pt_vois6 - 1) << 5;
		if (xxx > 127 || xxx < 0) return(CG_NOT_BINARY);
		*pt_out = lut[xxx];
		if(*pt_in != lut[xxx]) *modif = 1;
	    }
	    pt_in++; pt_out++; pt_vois2++; pt_vois6++;
	}

	echange (haut, milieu, bas);
	(void) mm0_writim (out, j, buf_lines[3]);
    }
    return (RESULTAT_OK);
}

void cg_fillut (hex, lut, type)
int *hex, type;
PIX *lut;

{

    int i, j;

    switch (type) {
    case 1 : /* EROSION */
	for (i = 0; i < 128; i++) {
	    lut[i] = 1;
	    for (j = 0; j < 7; j++) {
		if (hex[j] != 2)
		    if (hex[j] != ((i >> j) & 1)) lut[i] = 0;
	    }
	}
	break;
    case 2 : /* DILATATION */
	for (i = 0; i < 128; i++) {
	    lut[i] = 0;
	    for (j = 0; j < 7; j++) {
		if (hex[j] != 2)
		    if ((1 - hex[j]) != ((i >> j) & 1)) lut[i] = 1;
	    }
	}
	break;
    case 3 : /* RING */
	for (i = 0; i < 128; i++) {
	    lut[i] = 1;
	    for (j = 0; j < 7; j++) {
		if (hex[j] != 2)
		    if (hex[j] != ((i >> j) & 1)) lut[i] = 0;
	    }
	}
	break;
    case 4 : /* AMINCISSEMENT */
	for (i = 0; i < 128; i++) {
	    lut[i] = 0;
	    for (j = 0; j < 7; j++) {
		if (hex[j] != 2)
		    if (hex[j] != ((i >> j) & 1)) lut[i] = i & 1;
	    }
	}
	break;
    case 5 : /* EPAISSISSEMENT */
	for (i = 0; i < 128; i++) {
	    lut[i] = 1;
	    for (j = 0; j < 7; j++) {
		if (hex[j] != 2)
		    if (hex[j] != ((i >> j) & 1)) lut[i] = i & 1;
	    }
	}
	break;
    }

}

void cg_rotation (elt)
int *elt;
{
    int temp;

    temp = elt[1];
    elt[1] = elt[2]; elt[2] = elt[3]; elt[3] = elt[4];
    elt[4] = elt[5]; elt[5] = elt[6]; elt[6] = temp;
}

/* les scories */

void mm0_readim (im, y, pt)
     IMAGE *im;
     int  y;
     PIX *pt;
{
    PIX *ptend, *ptim;
    
    /* more efficient implementation to use bcopy */
    pt += im->lx - 1;
    ptend = pt + im->sx + 2;
    ptim  = im->bitmap + im->lx - 1 + y*im->tx;
    
    while (pt < ptend) *pt++ = *ptim++;
}


void mm0_writim (im, y, pt)
     IMAGE *im;
     int y;
     PIX *pt;
{
    PIX *ptend, *ptim;

    pt += im->lx - 1;
    ptend = pt + im->sx + 2;
    ptim  = im->bitmap + im->lx - 1 + y*im->tx;

    while (pt < ptend) *ptim++ = *pt++;

}

#define NBUFL 5			/* line buffers */

static int size=0;
static PIX **lines=0;

void mm0_free_lbuf()		/*  free line buffers */
{
    int  i;
    PIX *temp;

    /* deallocate previous buffers */
    if(lines!=NULL) {
	for (i = 0; i < NBUFL; i++) {
	    if((temp=lines[i])!=NULL) free(temp);
	}
	free(lines);
    }
    size=0;
    lines=0;
}

PIX **mm0_alloc_lbuf ( image)
     IMAGE *image;
{
    int  i, j;
    PIX  *temp;
    
    if ((lines!=0)&&(size >= image->tx)) {
	return(lines);
    }
    mm0_free_lbuf();
    /* allocate anew: calloc gives NULL space */
    if ((lines = (PIX **) calloc(NBUFL, sizeof(PIX *))) == NULL) {
	return(NULL);
    }
    size=image->tx;		/* this current image size */
    for (i = 0; i < NBUFL; i++) {
	if ((lines[i] =
	     (PIX *) malloc(sizeof(PIX)*(size + 4))) == NULL) {
	    mm0_free_lbuf();	/* error, free earlier allocation */
	    return(NULL);
	}
    }
    
    for (i = 0; i < NBUFL; i++) {
	for (j = 0; j < size + 4; j++) {
	    lines[i][j] = 1;
	}
    }
    return (lines);
}
#undef NBUFL

/*
 * squelette
 *
 * (de *squelette (in out)
 *     (*hmtbin in out #[2 1 1 2 0 0 2] 0 0 4))
 */

int	 mm0_squelette ( im_in, im_out)
     IMAGE *im_in;
     IMAGE *im_out;
{
    static int elt[7]={2, 1, 1, 2, 0, 0, 2};
    
    return ( mm0_hmtbin(im_in, im_out, elt, 0, 0, 4));
}
