/* basic io routines to test mm0 level */

#include "morpho.h"
#include "inr.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


static unsigned char tampon[2402]; /*  not clean but code will be dropped */

int mm0_imagin (im, nom)	/* lecture d'un fichier inrimage */
     IMAGE *im;
     char *nom;
{
    int fd, j, n, i;
    PIX *pt;
    struct descripteur header;
    
    if ((fd = open (nom, 0)) == -1) {
	return (OPEN_FAILED);
    }

    read (fd, &header, sizeof (struct descripteur));

    if (header.ndimv != 1) {
	return (BAD_FILE_FORMAT);
    }

    if ((im->lx + header.dimx > im->tx) || (im->ly + header.ndimy > im->ty)) {
	return (FILE_MEMORY_MISMATCH);
    }

    im->sx = header.dimx;
    im->sy = header.ndimy;

    if (header.flag == 0 && header.bsize == 1) {
	for (j = im->ly; j < im->ly + im->sy; j++) {
#ifdef PIX_UCHAR
	    if (read (fd, im->bitmap + im->lx + j*im->sx, im->sx) != im->sx) {
		close (fd);
		return (READ_FAILED);
	    }
#else
	    if ((n = read (fd, tampon, im->sx)) != im->sx) {
		fprintf(stderr," ** read error at line %d, %d bytes read\n\r",
			j, n);
		close (fd);
		return (READ_FAILED);
	    }
	    pt = im->bitmap + im->lx + j*im->tx;
	    for (i = 0; i < im->sx; i++)  *pt++ = (PIX) tampon[i];
#endif
	}
    }

    close (fd);
    return (RESULTAT_OK);
}

int mm0_imaout (im, nom)	/* mise en fichier inrimage */
     IMAGE *im;
     char *nom;
{
    
    int  i, j, n, ret;
    FILE *fd;
    unsigned char *pt_tampon;
    struct descripteur header;
    PIX *pt;
    
    if ((ret = BASIC_CHECK (im)) != RESULTAT_OK) return (ret);
    
    if ((fd = fopen (nom, "w")) == NULL) {
	return (OPEN_FAILED);
    }
    
    header.dimx = im->sx;
    header.ndimy = im->sy;
    header.bsize = 1;
    header.flag = 0;
    header.ndimz = 1;
    header.ndimv = 1;
    header.ndimx = im->sx;
    header.maximum = header.minimum = header.moyenne = 0;
    header.exposant = 0;
    header.float_typ = 0;
    header.im_magic = IM3_MAGIC;
    header.desc_ext = 0;
    header.x_off = header.y_off = header.z_off = 0;
    header.maille = 1;
    header.bias = header.scale = 0;
    
    if(fwrite (&header, sizeof (struct descripteur), 1, fd) != 1) {
	return (WRITE_FAILED);
    }
    
    for (j = im->ly; j < im->ly + im->sy; j++) {
#ifdef PIX_UCHAR
	if ((n = fwrite (im->bitmap + im->lx + j*im->sx, sizeof(PIX),
			 im->sx, fd)) != im->sx) {
	    fclose (fd);
	    return (WRITE_FAILED);
	}
#else
	pt = im->bitmap + im->lx + j*im->tx;
	pt_tampon = tampon;
	/* transform short into unsigned char ***** not so good  */
	for (i = 0; i < im->sx; i++) *pt_tampon++ = *pt++;
	if ((n = fwrite (tampon, sizeof(unsigned char), im->sx, fd))
	    != im->sx) {
	    perror("in imaout");
	    fprintf(stderr,"\r** fwrite error at line %d, %d bytes written\n\r",
		    j, n);
	    fclose (fd);
	    return (WRITE_FAILED);
	}
#endif
    }
    
    fclose (fd);
    return (RESULTAT_OK);
}

/*******************************/
/* lecture au format image ppm */
/*******************************/

int mm0_imagin_ppm (im, nom)
IMAGE * im;
char *nom;
{
  FILE *fp;
  int sx, sy, i, j, junk;
  char chaine_caract[1000];

  if ((fp = fopen (nom, "r")) == NULL) {
    fprintf (stderr, " ** erreur ouverture de %s\n", nom);
    exit (-1);
  }

  fgets (chaine_caract, 1000, fp);
  if (strcmp (chaine_caract, "P2\n") != 0) {
    fprintf (stderr, " ** Mauvais type de fichier (bye)\n\r");
    exit (-1);
  }

  do {
    fgets (chaine_caract, 1000, fp);
  } while (chaine_caract[0] == '#');

  sscanf (chaine_caract, "%d%d", &sx, &sy);
  printf ("tailles : %d %d\n", sx, sy);
  fgets (chaine_caract, 1000, fp);
  if (strcmp (chaine_caract, "255\n") != 0) {
    fprintf (stderr, " ** Pas 8 bits par point (bye)\n\r");
    exit (-1);
  }

  if (im->tx < sx + 2 || im->ty < sy + 2) {
    fprintf (stderr, " *** image allouee insuffisante\n");
    exit (-1);
  }

  im->sx = sx; im->sy = sy;
  im->lx = 1; im->ly = 1;

  /* lecture de l'image */

  for (j = 1; j <= sy; j++) {
    for (i = 1; i <= sx; i++) {
      fscanf (fp, "%d", &junk);
      im->bitmap[i + j*im->tx] = (PIX) junk;
    }
  }

  close (fp);

  return (RESULTAT_OK);
}

/*******************************/
/* lecture au format image pgm */
/*******************************/

int mm0_imagin_pgm (im, nom)
IMAGE * im;
char *nom;
{
  FILE *fp;
  int sx, sy, i, j;
  char chaine_caract[1000];

  if ((fp = fopen (nom, "r")) == NULL) {
    fprintf (stderr, " ** erreur ouverture de %s\n", nom);
    exit (-1);
  }

  fgets (chaine_caract, 1000, fp);
  if (strcmp (chaine_caract, "P5\n") != 0) {
    fprintf (stderr, " ** Mauvais type de fichier (bye)\n\r");
    exit (-1);
  }

  do {
    fgets (chaine_caract, 1000, fp);
  } while (chaine_caract[0] == '#');

  sscanf (chaine_caract, "%d%d", &sx, &sy);
  printf ("tailles : %d %d\n", sx, sy);
  fgets (chaine_caract, 1000, fp);
  if (strcmp (chaine_caract, "255\n") != 0) {
    fprintf (stderr, " ** Pas 8 bits par point (bye)\n\r");
    exit (-1);
  }

  if (im->tx < sx + 2 || im->ty < sy + 2) {
    fprintf (stderr, " *** image allouee insuffisante\n");
    exit (-1);
  }

  im->sx = sx; im->sy = sy;
  im->lx = 1; im->ly = 1;

  /* lecture de l'image */

  for (j = 1; j <= sy; j++) {
    for (i = 1; i <= sx; i++) {
      im->bitmap[i + j*im->tx] = (PIX) (getc (fp));
    }
  }

  close (fp);

  return (RESULTAT_OK);
}

int mm0_imaout_pgm (im, nom)
IMAGE * im;
char *nom;
{
  FILE *fp;
  int i, j;
  char chaine_caract[1000];

  if ((fp = fopen (nom, "w")) == NULL) {
    fprintf (stderr, " ** erreur ouverture de %s\n", nom);
    exit (-1);
  }

  fputs ("P5\n", fp);

  sprintf (chaine_caract, "%4d %4d\n", im->sx, im->sy);
  printf ("tailles : %d %d\n", im->sx, im->sy);
  fputs (chaine_caract, fp);
  fputs ("255\n", fp);

  /* ecriture de l'image */

  for (j = im->ly; j < im->sy + im->ly; j++) {
    for (i = im->lx; i < im->sx + im->lx; i++) {
      putc ((int) im->bitmap[i + j*im->tx], fp);
    }
  }

  close (fp);

  return (RESULTAT_OK);
}

int mm0_imagin_pgm_ascii (im, nom)
IMAGE * im;
char *nom;
{
  FILE *fp;
  int sx, sy, i, j, value;
  char chaine_caract[1000];

  if ((fp = fopen (nom, "r")) == NULL) {
    fprintf (stderr, " ** erreur ouverture de %s\n", nom);
    exit (-1);
  }

  fgets (chaine_caract, 1000, fp);
  if (strcmp (chaine_caract, "P2\n") != 0) {
    fprintf (stderr, " ** Mauvais type de fichier (bye)\n\r");
    exit (-1);
  }

  do {
    fgets (chaine_caract, 1000, fp);
  } while (chaine_caract[0] == '#');

  sscanf (chaine_caract, "%d%d", &sx, &sy);
  printf ("tailles : %d %d\n", sx, sy);
  fgets (chaine_caract, 1000, fp);
  if (strcmp (chaine_caract, "255\n") != 0) {
    fprintf (stderr, " ** Pas 8 bits par point (bye)\n\r");
    exit (-1);
  }

  if (im->tx < sx + 2 || im->ty < sy + 2) {
    fprintf (stderr, " *** image allouee insuffisante\n");
    exit (-1);
  }

  im->sx = sx; im->sy = sy;
  im->lx = 1; im->ly = 1;

  /* lecture de l'image */

  for (j = 1; j <= sy; j++) {
    for (i = 1; i <= sx; i++) {
      fscanf (fp, "%d", &value);
      im->bitmap[i + j*im->tx] = value;
    }
  }

  close (fp);

  return (RESULTAT_OK);
}

int mm0_imagin_pgm_c (im, nom)
IMAGE * im;
char *nom;
{
  FILE *fp;
  int sx, sy, i, j;
  char chaine_caract[1000];

  if ((fp = fopen (nom, "r")) == NULL) {
    fprintf (stderr, " ** erreur ouverture de %s\n", nom);
    exit (-1);
  }

  fgets (chaine_caract, 1000, fp);
  if (strcmp (chaine_caract, "P6\n") != 0) {
    fprintf (stderr, " ** Mauvais type de fichier (bye)\n\r");
    exit (-1);
  }

  do {
    fgets (chaine_caract, 1000, fp);
  } while (chaine_caract[0] == '#');

  sscanf (chaine_caract, "%d%d", &sx, &sy);
  printf ("tailles : %d %d\n", sx, sy);
  fgets (chaine_caract, 1000, fp);
  if (strcmp (chaine_caract, "255\n") != 0) {
    fprintf (stderr, " ** Pas 8 bits par point (bye)\n\r");
    exit (-1);
  }

  if (im->tx < sx + 2 || im->ty < sy + 2) {
    fprintf (stderr, " *** image allouee insuffisante\n");
    exit (-1);
  }

  im->sx = sx; im->sy = sy;
  im->lx = 1; im->ly = 1;

  /* lecture de l'image */

  for (j = 1; j <= sy; j++) {
    for (i = 1; i <= sx; i++) {
      im->bitmap[i + j*im->tx] = (PIX) (getc (fp));
      getc (fp); getc (fp);
    }
  }

  close (fp);

  return (RESULTAT_OK);
}

int mm0_imagin_pgm_cc (im_r, im_g, im_b, nom)
IMAGE * im_r, *im_g, *im_b;
char *nom;
{
  FILE *fp;
  int sx, sy, i, j;
  char chaine_caract[1000];

  if ((fp = fopen (nom, "r")) == NULL) {
    fprintf (stderr, " ** erreur ouverture de %s\n", nom);
    exit (-1);
  }

  fgets (chaine_caract, 1000, fp);
  if (strcmp (chaine_caract, "P6\n") != 0) {
    fprintf (stderr, " ** Mauvais type de fichier (bye)\n\r");
    exit (-1);
  }

  do {
    fgets (chaine_caract, 1000, fp);
  } while (chaine_caract[0] == '#');

  sscanf (chaine_caract, "%d%d", &sx, &sy);
  printf ("tailles : %d %d\n", sx, sy);
  fgets (chaine_caract, 1000, fp);
  if (strcmp (chaine_caract, "255\n") != 0) {
    fprintf (stderr, " ** Pas 8 bits par point (bye)\n\r");
    exit (-1);
  }

  if (im_r->tx < sx + 2 || im_r->ty < sy + 2
      || im_g->tx < sx + 2 || im_g->ty < sy + 2
      || im_b->tx < sx + 2 || im_b->ty < sy + 2) {
    fprintf (stderr, " *** image allouee insuffisante\n");
    exit (-1);
  }

  im_r->sx = im_g->sx = im_b->sx = sx;
  im_r->sy = im_g->sy = im_b->sy = sy;
  im_r->lx = im_g->lx = im_b->lx = 1;
  im_r->ly = im_g->ly = im_b->ly = 1;

  /* lecture de l'image */

  for (j = 1; j <= sy; j++) {
    for (i = 1; i <= sx; i++) {
      im_r->bitmap[i + j*im_r->tx] = (PIX) (getc (fp));
      im_g->bitmap[i + j*im_g->tx] = (PIX) (getc (fp));
      im_b->bitmap[i + j*im_b->tx] = (PIX) (getc (fp));
    }
  }

  close (fp);

  return (RESULTAT_OK);
}

int mm0_imaout_pgm_cc (im_r, im_g, im_b, nom)
IMAGE * im_r, *im_g, *im_b;
char *nom;
{
  FILE *fp;
  int i, j;
  char chaine_caract[1000];

  if ((fp = fopen (nom, "w")) == NULL) {
    fprintf (stderr, " ** erreur ouverture de %s\n", nom);
    exit (-1);
  }

  fputs ("P6\n", fp);

  sprintf (chaine_caract, "%4d %4d\n", im_r->sx, im_r->sy);
  printf ("tailles : %d %d\n", im_r->sx, im_r->sy);
  fputs (chaine_caract, fp);
  fputs ("255\n", fp);

  /* ecriture de l'image */

  for (j = im_r->lx; j <= im_r->sy; j++) {
    for (i = im_r->ly; i <= im_r->sx; i++) {
      putc ((int) im_r->bitmap[i + j*im_r->tx], fp);
      putc ((int) im_g->bitmap[i + j*im_r->tx], fp);
      putc ((int) im_b->bitmap[i + j*im_r->tx], fp);
    }
  }

  close (fp);

  return (RESULTAT_OK);
}

/***************************/
/* lecture au format image */
/***************************/

int mm0_imagin_enst (im, nom)
IMAGE * im;
char *nom;
{
  FILE *fp;
  int fd;
  int sx, sy, i, j;
  unsigned char donnee;
  char nom_donnee[1000], nom_taille[1000];

  sprintf (nom_donnee, "%s.ima", nom);
  sprintf (nom_taille, "%s.dim", nom);

  if ((fp = fopen (nom_taille, "r")) == NULL) {
    fprintf (stderr, " ** erreur ouverture de %s\n", nom_taille);
    exit (-1);
  }

  fscanf (fp, "%d%d", &sx, &sy);
  fclose (fp);

  printf ("tailles : %d %d\n", sx, sy);

  if (im->tx < sx + 2 || im->ty < sy + 2) {
    fprintf (stderr, " *** image allouee insuffisante\n");
    exit (-1);
  }

  im->sx = sx; im->sy = sy;
  im->lx = 1; im->ly = 1;

  /* lecture de l'image */

  if ((fd = open (nom_donnee, O_RDONLY,0)) == -1) {
    fprintf (stderr, " *** erreur ouverture de %s\n", nom_donnee);
    exit (-1);
  }

  for (j = 1; j <= sy; j++) {
    for (i = 1; i <= sx; i++) {
      read (fd, &donnee, sizeof (unsigned char));
      im->bitmap[i + j*im->tx] = (PIX) donnee;
    }
  }

  close (fd);

  return (RESULTAT_OK);
}

int mm0_imaout_enst (im, nom)
IMAGE * im;
char *nom;
{
  FILE *fp;
  int fd;
  int i, j;
  unsigned char donnee;
  char nom_donnee[1000], nom_taille[1000];

  sprintf (nom_donnee, "%s.ima", nom);
  sprintf (nom_taille, "%s.dim", nom);

  if ((fp = fopen (nom_taille, "w")) == NULL) {
    fprintf (stderr, " ** erreur ouverture de %s\n", nom_taille);
    exit (-1);
  }

  fprintf (fp, "%d %d", im->sx, im->sy);
  fclose (fp);

  printf ("tailles : %d %d\n", im->sx, im->sy);

  /* ecriture de l'image */

  fp = fopen (nom_donnee, "w"); fclose (fp);
  if ((fd = open (nom_donnee, O_WRONLY, 0)) == -1) {
    fprintf (stderr, " *** erreur ouverture de %s\n", nom_donnee);
    exit (-1);
  }

  for (j = 1; j <= im->sy; j++) {
    for (i = 1; i <= im->sx; i++) {
      donnee = (unsigned char) im->bitmap[i + j*im->tx];
      write (fd, &donnee, sizeof (unsigned char));
    }
  }

  close (fd);

  return (RESULTAT_OK);
}
