#include "morpho.h"
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <malloc.h>


/***************************/
/* error handling routines */
/***************************/

void terror(val,text1,text2)
int  val;
char *text1, *text2;
{
  char *e_cg();

  printf("*** error %s [code=%d] in '%s' at '%s'\n",
	 e_cg(val),val,text1,text2);
  exit(-1);

}

void killer()
{
  abort();
}

/*****************************/
/* initialisation des images */
/*****************************/

IMAGE ** init_all_images (nb_im, sx, sy)
int nb_im, sx, sy;
{
  IMAGE ** liste;
  int i;

  liste = (IMAGE **) malloc (nb_im*sizeof(IMAGE *));
  if (liste == NULL) {
    fprintf (stderr, " *** Erreur d'allocation\n");
    exit (-1);
  }

  for (i = 0; i < nb_im; i++) {
    liste[i] = (IMAGE *) mm0_alloc_image (sx + 2, sy + 2, 1, 1, sx, sy);
    if(liste[i]==NULL) {
      fprintf (stderr, " *** Erreur d'allocation\n");
      exit (-1);
    }
  }

  return (liste);
}

IMAGE *mm0_alloc_image (tx, ty, lx, ly, sx, sy)
     int tx, ty, lx, ly, sx, sy;
{
    IMAGE *im = NULL, *mm0_alloc_header();
    
    /* test de consistance */
    
    if ((lx + sx > tx) || (ly + sy > ty)) {
	return (NULL);
    }
    
    /* creation du header */
    
    im = mm0_alloc_header ();
    
    /* creation du bitmap */
    
    if (mm0_alloc_bitmap (im, tx, ty) != RESULTAT_OK) {
	free ((char *) im);
	return (NULL);
    }
    
    im->lx = lx; im->ly = ly;
    im->sx = sx; im->sy = sy;
    im->tx = tx; im->ty = ty;
    
    return (im);
}

IMAGE *mm0_alloc_header ()
{
    IMAGE *im;
    
    im = (IMAGE *) malloc (sizeof(IMAGE));
    if (im == NULL) return (NULL);
    
    im->tx = im->ty = 0;
    im->lx = im->ly = im->sx = im->sy = 0;
    return (im);
}

int mm0_alloc_bitmap (im, tx, ty)
     IMAGE *im;
     int tx, ty;
{
    im->bitmap = (PIX *) malloc (tx * ty * sizeof(PIX));
    if (im->bitmap == NULL) return (ALLOC_FAILED);
    
    im->tx = tx;
    im->ty = ty;
    return (RESULTAT_OK);
}

int mm0_free_bitmap (im)
     IMAGE *im;
{
    int ret;
    
    (void) free((char *) im->bitmap);
    
    im->sx = im->sy = 0;
    im->bitmap = NULL;
    return (RESULTAT_OK);
}

int mm0_print_caract (im)
     IMAGE *im;
{
    printf ("%x : (taille physique [%d %d] lup [%d %d] size[%d %d] bitmap[%x])\n\r",
	    im,
	    im->sx, im->sy,
	    im->lx, im->ly,
	    im->sx, im->sy,
	    im->bitmap);
    fflush (stdout);
    
    return (0);
}

int mm0_modif_caract (im, lx, ly, sx, sy)
     IMAGE *im;
     int lx, ly, sx, sy;
{
    if ((lx + sx > im->sx) || (ly + sy > im->sy))
      return (IMAGE_TOO_SMALL);
    
    im->lx = lx; im->ly = ly;
    im->sx = sx; im->sy = sy;
    
    return (0);
}

typedef struct { int code; char act; char *text;} erec;
static erec ecodes[]={
	{100,'k',"parameter 1 is not an allocated image"},
	{101,'k',"parameter 2 is not an allocated image"},
	{102,'k',"parameter 3 is not an allocated image"},
	{103,'k',"parameter 4 is not an allocated image"},
	{104,'k',"empty intersection on X (xmax < xmin)"},
	{105,'k',"empty intersection on Y (ymax < ymin)"},
	{106,'k',"integer parameter not in allowed range"},
	{107,'k',"conversion ipl->inr cannot fit in allocated max_size"},
	{108,'k',"conversion ipl->inr pixel type not supported"},
	{120,'k',"failed MALLOC"},
	{121,'k',"Pixel value not in range"},
	{122,'k',"Pixel value must not be negative"},
	{123,'k',"the 2 images MUST be different"},
	{124,'k',"the input image MUST be binary: pixel values 0 or 1"},
	{125,'k',"direction must be a multiple of 30 degrees"},
	{126,'k',"cannot allocate temporary image buffer"},
	{127,'k',"could not find such pixel value in image"},
	{128,'k',"error"},
	{129,'k',"floating parameter not in allowed range"},
	{151,'k',"mm0: fopen failed"},
	{152,'k',"mm0: bad_file_format"},
	{153,'k',"mm0: file_memory_mismatch"},
	{154,'k',"mm0: read_failed"},
	{155,'k',"mm0: write_failed"},
	{156,'k',"mm0: image_too_small"},
	{157,'k',"mm0: image_not_different"},
	{158,'k',"mm0: image_not_same_size"},
	{159,'k',"mm0: param_not_in_range"},
	{160,'k',"mm0: alloc_failed"},
	{161,'k',"mm0: null_header"},
	{162,'k',"mm0: null_bitmap"},
	{163,'k',"mm0: incoherent_header"},
	{164,'k',"mm0: free_failed"},
	{165,'k',"mm0: cannot fit frame in X"},
	{166,'k',"mm0: cannot fit frame in Y"},
	{167,'k',"mm0: coordinate X not in window"},
	{168,'k',"mm0: coordinate Y not in window"},
	{169,'k',"mm0: alloc_lacet failed"},
	{170,'k',"mm0: images not allocated same in X"},
	{171,'k',"dil_chaine_conj: can't dilate single point"},
	{172,'k',"mm0: cre_elem malloc failed"}};

#define cg_nerr 41

char *e_cg(num)
	int num;
{
	int i;

	num &= 0xffff;	/* keep lower half */

	for (i=0; i < cg_nerr; i++) {
		if(ecodes[i].code==num) {
			return ( ecodes[i].text);
		}
	}
	return "e_cg unknown error code";
}

int cg_error(num, fun)
	int num; char *fun;
{
	int i; char *etext;

	if(fun==0) fun="****";
	etext = e_cg(num);
	fprintf(stderr,"%s in %s\n\r",etext,fun);
	if( ecodes[i].act =='k') killer();

	return( -1);
}


/* "global" memory kept local to operate lacet allocations */
static int nb_dir=0;
static int nb_bou=0;
static GLACE cg_lacet;

/* prepare new set of lacet operations */
void mm0_reset_lacet()
{
    *cg_lacet.chain1 = 1;
    *cg_lacet.chain2 = 1;

    cg_lacet.pt = 0;
    cg_lacet.pt1 = 0;
    cg_lacet.pt2 = 0;
    cg_lacet.dir = 0;
    cg_lacet.imavide = 0;
    (cg_lacet.loop1)->origine.x = 1;
    (cg_lacet.loop2)->origine.x = 1;
    cg_lacet.xx = 0;
    cg_lacet.yy = 0;
}

void mm0_free_lacet()
{
    if(cg_lacet.chain1!=NULL) {
	free(cg_lacet.chain1);
	cg_lacet.chain1=NULL;
    }
    if(cg_lacet.chain2!=NULL) {
	free(cg_lacet.chain2);
	cg_lacet.chain2=NULL;
    }
    if(cg_lacet.loop1!=NULL) {
	free(cg_lacet.loop1);
	cg_lacet.loop1=NULL;
    }
    if(cg_lacet.loop2!=NULL) {
	free(cg_lacet.loop2);
	cg_lacet.loop2=NULL;
    }
    nb_bou=0;
    nb_dir=0;
}

GLACE *mm0_alloc_lacet (nb_direction,nb_boucles)
     long nb_direction,nb_boucles;
{
    /* allocates if necessary the loops & chains space, sets up the 
     * starting values for loops and chains.
     * if failed, returns NULL, else pointer to GLACE loops&chain struct
     */
    int alloc=0;
    size_t asize;

    if((nb_dir==0)||(nb_bou==0)) {
	alloc=1;
    } else {
	/* reallocate storage if more space required */
	if((nb_direction>nb_dir)||(nb_boucles>nb_bou)) {
	    (void)mm0_free_lacet();
	    alloc=1;
	}
    }
    if(alloc!=0) {
	asize = sizeof(DIRECTION)*nb_direction;
	/* printf("mm0_alloc_lacet chain1 size=%d\n",asize); fflush(stdout); */
 	if ((cg_lacet.chain1 = (DIRECTION *) malloc(asize)) == NULL) {
	    /* " ** Impossible d'allouer cg_lacet.chain1\n\r"); */
	    return(NULL);
	}
	/* printf("mm0_alloc_lacet chain2\n"); fflush(stdout); */
	if ((cg_lacet.chain2 = (DIRECTION *) malloc(asize)) == NULL) {
	    /* " ** Impossible d'allouer cg_lacet.chain2\n\r"); */
	    (void)mm0_free_lacet();
	    return(NULL);
	}
	asize = sizeof(struct lacet)*nb_boucles;
	/* printf("mm0_alloc_lacet loop1\n");     fflush(stdout); */
	if ((cg_lacet.loop1 = (struct lacet *) malloc(asize))==NULL) {
	    /* " ** Impossible d'allouer cg_lacet.loop1\n\r"); */
	    (void)mm0_free_lacet();
	    return(NULL);
	}
	/* printf("mm0_alloc_lacet loop2\n"); fflush(stdout); */
	if ((cg_lacet.loop2 = (struct lacet *) malloc(asize)) ==NULL) {
	    /* " ** Impossible d'allouer cg_lacet.loop2\n\r"); */
	    (void)mm0_free_lacet();
	    return(NULL);
	}
    }
    /* initialise all variables */
    (void)mm0_reset_lacet();
    nb_bou=nb_boucles;
    nb_dir=nb_direction;
    return (&cg_lacet);
}

/* simpler routine to allocate loops & chains */
GLACE *mm0_alloc_chains (image)
     IMAGE *image;
{
    long   nb_direction,nb_boucles;
    GLACE *cgl;

    nb_direction=image->tx;
    nb_direction *=image->ty;
    nb_direction /=2;
    nb_boucles=image->tx;
    nb_boucles*=image->ty;
    nb_boucles /=10;
    return (mm0_alloc_lacet (nb_direction,nb_boucles));
}

/* ! THIS ROUTINE must not be called by user codes !!*/
/*****************************************************/
/* regles de reecriture pour l'hexagone de taille un */
/*****************************************************/

void mm0_recri_dil (cgl)
     GLACE *cgl;
{
    int dif;

    dif = (*cgl->pt2 - *cgl->pt1); if (dif < 0) dif += 6;
    switch (dif)
	{
	case 0 : *cgl->pt++ = *cgl->pt1;
		 break;
        case 1 : *cgl->pt++ = *cgl->pt1;
	         *cgl->pt++ = *cgl->pt2;
	         break;
        case 2 : *cgl->pt++ = *cgl->pt1;
	         *cgl->pt++ = (*cgl->pt1 < 5) ? *cgl->pt1 + 1 : *cgl->pt1 - 5;
	         *cgl->pt++ = *cgl->pt2;
                 break;
       case 3 : *cgl->pt++ = *cgl->pt1;
 	        *cgl->pt++ = (*cgl->pt1 < 5) ? *cgl->pt1 + 1 : *cgl->pt1 - 5;
	        *cgl->pt++ = (*cgl->pt1 < 4) ? *cgl->pt1 + 2 : *cgl->pt1 - 4;
	        *cgl->pt++ = *cgl->pt2;
	        break;
       case 4 : *cgl->pt++ = *cgl->pt1;
	        *cgl->pt++ = (*cgl->pt1 < 5) ? *cgl->pt1 + 1 : *cgl->pt1 - 5;
	        *cgl->pt++ = (*cgl->pt1 < 4) ? *cgl->pt1 + 2 : *cgl->pt1 - 4;
	        *cgl->pt++ = (*cgl->pt1 < 3) ? *cgl->pt1 + 3 : *cgl->pt1 - 3;
	        *cgl->pt++ = *cgl->pt2;
	        break;
	case 5 : break;
	     }
}

/* ! THIS ROUTINE must not be called by user codes !!*/
int mm0_depla (cgl, image)
     GLACE *cgl;
     IMAGE *image;
{
    
    /***************************************************************/
    /* calcule la difference d'adresse dans la direction cgl->dir */
    /* on utilise les variables cgl->xx et cgl->yy donnant       */
    /* la position du point courrant                               */
    /***************************************************************/
    
    switch (cgl->dir) {
	case 0 : return (1); break;
	case 1 : return (image->tx + (cgl->yy & 1)); break;
	case 2 : return (image->tx - 1 + (cgl->yy & 1)); break;
	case 3 : return (-1); break;
	case 4 : return (-image->tx - 1 + (cgl->yy & 1)); break;
	case 5 : return (-image->tx +  (cgl->yy & 1)); break;
    }
    abort();			/* never used: trap it! */
}

/* ! THIS ROUTINE must not be called by user codes !!*/
void mm0_xynew ( cgl)
     GLACE *cgl;
{

/***************************************************************/
/* met a jour la position courrante cgl->xx et cgl->yy quand on  */
/* on se deplace dans la direction cgl->dir                     */
/***************************************************************/

    switch (cgl->dir) {
    case 0 : cgl->xx++; break;
    case 1 : cgl->xx += (cgl->yy & 1); cgl->yy++; break;
    case 2 : cgl->xx += - 1 + (cgl->yy & 1); cgl->yy++; break;
    case 3 : cgl->xx--; break;
    case 4 : cgl->xx += - 1 + (cgl->yy & 1); cgl->yy--; break;
    case 5 : cgl->xx += (cgl->yy & 1); cgl->yy--; break;
	 }
}

/* ! THIS ROUTINE must not be called by user codes !!*/
/*****************************************************/
/* dilatation par un hexagone de taille un de toutes */
/* les chaines de cg_loop1 dans cg_loop2           */
/* cg_loop1 et cg_chain1 restent intacts           */
/*****************************************************/

void mm0_dil_chaine (cgl)
     GLACE *cgl;
{
    struct lacet *boucle1, *boucle2;
    int i;

    boucle1 = cgl->loop1;
    boucle2 = cgl->loop2;
    cgl->pt = cgl->chain2;

    while (boucle1->liste != 0) {

	/* cas du point isole */

	if (boucle1->longueur == 0) {
	    boucle2->origine.x = boucle1->origine.x + 1;
	    boucle2->origine.y = boucle1->origine.y;
	    boucle2->valeur = boucle1->valeur;
	    boucle2->liste = cgl->pt;
	    boucle2->longueur = 6;
	    boucle2->type = 1;
	    *cgl->pt++ = 2; *cgl->pt++ = 3; *cgl->pt++ = 4;
	    *cgl->pt++ = 5; *cgl->pt++ = 0; *cgl->pt++ = 1;
	} else {
	    boucle2->liste = cgl->pt; boucle2->type = boucle1->type;
	    boucle2->valeur = boucle1->valeur;
	    if (boucle1->type == 1) {

		/* cas du lacet qui ne doit pas arriver !! */

		cgl->pt1 = boucle1->liste + boucle1->longueur - 1;
		cgl->pt2 = boucle1->liste;
		cgl->xx = boucle1->origine.x;
		cgl->yy = boucle1->origine.y;
		cgl->dir = (*cgl->pt1 < 2) ?
		  *cgl->pt1 + 4 : *cgl->pt1 - 2; mm0_xynew (cgl);
		boucle2->origine.x = cgl->xx;
		boucle2->origine.y = cgl->yy;
		for (i = 0; i < boucle1->longueur; i++) {
		    mm0_recri_dil (cgl);
		    cgl->pt1 = cgl->pt2; cgl->pt2++;
		}
	    } else {

		/* cas general d'une chaine non vide */

		cgl->pt1 = boucle1->liste;
		cgl->pt2 = boucle1->liste + 1;
		boucle2->origine.x = boucle1->origine.x;
		boucle2->origine.y = boucle1->origine.y;
		*cgl->pt++ = (*cgl->pt1 < 1) ?
		  *cgl->pt1 + 5 : *cgl->pt1 - 1;
		for (i =0; i < boucle1->longueur - 1; i++) {
		    mm0_recri_dil (cgl);
		    cgl->pt1 = cgl->pt2; cgl->pt2++;
		}
		*cgl->pt++ = (*cgl->pt1 < 5) ?
		  *cgl->pt1 + 1 : *cgl->pt1 - 5;
	    }
	}
	boucle2->longueur = cgl->pt - boucle2->liste;
	boucle2++; boucle1++;
    }
    boucle2->liste = 0;
}

/*******************************************************/
/* ajustement des chaines et lacets de cg_loop2       */
/* resultat dans cg_loop1                             */
/* im : image contenant le masque de propagation       */
/* val_trace : tous les lacets et chaines sont traces  */
/*       a la valeur val_trace                         */
/* val_masque : seules les portions de chaines qui se  */
/*       trouvaient sur des pixels a cette valeur sont */
/* conservees.                                         */
/*******************************************************/

void mm0_adjust_masque (image, cgl, val_trace, val_masque)
     IMAGE *image;
     GLACE *cgl;
     int val_trace, val_masque;
{
    struct lacet *bouc1, *bouc2;
    PIX c, *lecture;
    DIRECTION *ptend, *ptold, *pt1_temp, *pt2_temp;

    /* input checks are suppressed: they should have been done
     * by the caller. This routine is PRIVATE and is called
     * from routines having already checked their input
     */
    
    bouc2 = cgl->loop2; bouc1 = cgl->loop1;
    cgl->pt2 = cgl->chain2; cgl->pt1 = cgl->chain1;
    cgl->imavide = 1;
    
    while (bouc2->liste != 0) {
	ptend = bouc2->liste + bouc2->longueur;
	cgl->xx = bouc2->origine.x; cgl->yy = bouc2->origine.y;
	lecture = image->bitmap + cgl->xx + cgl->yy*image->tx;
	c = (val_trace < 0) ? bouc2->valeur : val_trace;
	/* printf ("[adjm:%d]", c); fflush (stdout); */
	cgl->pt2 = bouc2->liste;
	if (bouc2->type == 1) {
	    
	    /* cas d'un lacet */
	    
	    while (cgl->pt2 < ptend) {
		if (*lecture != val_masque) {
		    /* rotation du lacet -> chaine */
		    pt2_temp = bouc2->liste; pt1_temp = cgl->pt1;
		    while (pt2_temp < cgl->pt2) *pt1_temp++ = *pt2_temp++;
		    pt1_temp = bouc2->liste;
		    while (pt2_temp < ptend) *pt1_temp++ = *pt2_temp++;
		    pt2_temp = cgl->pt1;
		    while (pt1_temp < ptend) *pt1_temp++ = *pt2_temp++;
		    bouc2->origine.x = cgl->xx; bouc2->origine.y = cgl->yy;
		    bouc2->type = 0; break;
		}
		cgl->dir = *cgl->pt2++;
		lecture += mm0_depla (cgl,image); mm0_xynew (cgl);
	    }
	    if (bouc2->type == 1) {
		/* si lacet -> recopie et trace */
		cgl->xx = bouc1->origine.x = bouc2->origine.x;
		cgl->yy = bouc1->origine.y = bouc2->origine.y;
		bouc1->longueur = bouc2->longueur;
		bouc1->valeur = bouc2->valeur;
		bouc1->type = 1; bouc1->liste = cgl->pt1;
		lecture = image->bitmap + cgl->xx + cgl->yy*image->tx;
		*lecture = c; /* corrige le 24/4/98 */
		for(cgl->pt2 = bouc2->liste; cgl->pt2 < ptend;) {
		    cgl->dir = *cgl->pt2; 
		    lecture += mm0_depla (cgl, image);
		    *lecture = c; mm0_xynew (cgl);
		    *cgl->pt1++ = *cgl->pt2++;
		}
		bouc2++; bouc1++; cgl->imavide = 0;
	    }
	} else {
	    
	    /* cas de la chaine */
	    
	    for (;;) {
		cgl->dir = *cgl->pt2;
		/* recherche de debut de chaine */
		while(cgl->pt2 < 
		      ptend && *(lecture += mm0_depla (cgl,image)) != val_masque) {
		    mm0_xynew (cgl); cgl->pt2++; cgl->dir = *cgl->pt2;
		}
		bouc1->longueur = 0;
		if (cgl->pt2 == ptend) break;
		*lecture = c;
		bouc1->origine.x = cgl->xx; bouc1->origine.y = cgl->yy;
		bouc1->type = 0; bouc1->liste = cgl->pt1;
		ptold = cgl->pt2; mm0_xynew (cgl); *cgl->pt1++ = *cgl->pt2++;
		cgl->dir = *cgl->pt2;
		/* suivi de la chaine nouvelle */
		while(cgl->pt2 < ptend &&
		      *(lecture += mm0_depla (cgl,image)) == val_masque) {
		    *cgl->pt1++ = *cgl->pt2; *lecture = c;
		    mm0_xynew (cgl); cgl->pt2++; cgl->dir = *cgl->pt2;
		}
		mm0_xynew (cgl); *cgl->pt1++ = *cgl->pt2++;
		bouc1->valeur = bouc2->valeur;
		bouc1->longueur = cgl->pt2 - ptold; bouc1++;
		cgl->imavide = 0;
		if (cgl->pt2 == ptend) break;
	    }
	    bouc2++;
	}
    }
    bouc1->liste = 0;
}

int mm0_front (image, cgl, lx, ly, dirdep, boucle, ptdep)
     IMAGE     *image;
     int       lx, ly, dirdep;
     GLACE     *cgl;
     DIRECTION *ptdep;
     struct lacet *boucle;
{
    PIX *debut, *adr;
    cgl->pt1 = ptdep;
    cgl->xx = lx;
    cgl->yy = ly;
    cgl->dir = dirdep;
    debut = image->bitmap + cgl->xx + (cgl->yy)*(image->tx);
    *debut = 3; *(debut + mm0_depla (cgl, image)) = 2;
    if (++cgl->dir > 5) cgl->dir -= 6;

    while (1) {
      if ((cgl->xx != lx) || (cgl->yy != ly)) {
	/* cas du point general */
	while (!(*(adr = debut + mm0_depla (cgl, image)) & 1)) {
	  *adr = 2;
	  if (++cgl->dir > 5) cgl->dir -= 6;
	}

      } else {

	/* cas du point de depart */
	while (!(*(adr = debut + mm0_depla (cgl, image)) & 1)) {
	  if (cgl->dir == dirdep) {
	    /* fin de l'algorithme */
	    boucle->longueur = cgl->pt1 - ptdep;
	    boucle->origine.x = lx; boucle->origine.y = ly;
	    boucle->liste = ptdep; boucle->type = 1;
	    /* *debut = 3; */
	    return (RESULTAT_OK);
	  }
	  *adr = 2;
	  if (++cgl->dir > 5) cgl->dir -= 6;
	}
      }
      debut = adr;
      *cgl->pt1++ = cgl->dir;
      mm0_xynew (cgl);
      *debut = 3; cgl->dir += 4; if (cgl->dir > 5) cgl->dir -= 6;
    }
}

int mm0_front_old (image, cgl, lx, ly, dirdep, boucle, ptdep)
     IMAGE     *image;
     int       lx, ly, dirdep;
     GLACE     *cgl;
     DIRECTION *ptdep;
     struct lacet *boucle;
{
/********************************************************************/
/*    suivi d'une seule frontiere = une seule chaine                */
/*    im      : image de l'ensemble (fond = 0, forme = 1)           */
/*    lx,ly   : point de depart du suivi                            */
/*    dirdep  : direction de depart. Le point dans cette direction  */
/*              a partir de lx, ly doit etre dans le fond           */
/*    boucle  : entete de chaine ou stocker le resultat             */
/*    ptdep   : endroit ou stocker les directions                   */
/* !! im est modifiee : fond  = 0 ou 2 (erode du fond = 0)          */
/*                      forme = 1 ou 3 (erode de la forme = 3)      */
/* !! si le contour suivi est reduit a un point la chaine a une     */
/*    longueur 0                                                    */
/********************************************************************/

    PIX *debut, *adr;		/* apparently not initialised */
				/* but algorithm garantees not used before set */
    /* DIRECTION *pt1; */

    cgl->pt1 = ptdep;
    cgl->xx = lx;
    cgl->yy = ly;
    cgl->dir = dirdep + 1;
    debut = image->bitmap + cgl->xx + (cgl->yy)*(image->tx);

    while (1) {			/* forever */
	if ((cgl->xx != lx) || (cgl->yy != ly)) {
	    /* cas du point general */
	    do { if (++cgl->dir > 5) cgl->dir -= 6; if (debut != adr) *adr = 2;}
	    while (!(*(adr = debut + mm0_depla (cgl, image)) & 1));

	} else {

	    /* cas du point de depart */
	    if (cgl->dir > 5) cgl->dir -= 6;
	    while (!(*(adr = debut + mm0_depla (cgl, image)) & 1)) {
		*adr = 2;
		if (cgl->dir == dirdep) {
		    /* fin de l'algorithme */
		    boucle->longueur = cgl->pt1 - ptdep;
		    boucle->origine.x = lx; boucle->origine.y = ly;
		    boucle->liste = ptdep; boucle->type = 1;
		    *debut = 3;
		    return (RESULTAT_OK);
		}
		if (++cgl->dir > 5) cgl->dir -= 6;
	    }
	}
	debut = adr;
	*cgl->pt1++ = cgl->dir;
	mm0_xynew (cgl);
	*debut = 3; cgl->dir += 4;
    }
    /* never returning via here */
}

/* cette version est pour contours exterieurs */

mm0_front_e (image, cgl, lx, ly, dirdep, boucle, ptdep)
     int       lx, ly, dirdep;
     GLACE     *cgl;
     DIRECTION *ptdep;
     struct lacet *boucle;
     IMAGE     *image;
{
    PIX *debut, *adr;
    DIRECTION *cg_pt1;
    
    cgl->pt1 = ptdep;
    cgl->xx = lx; cgl->yy = ly; cgl->dir = dirdep + 1;
    debut = image->bitmap + cgl->xx + cgl->yy*image->tx;
    
    while (1) {
	
	if ((cgl->xx != lx) || (cgl->yy != ly)) {
	    
	    /* cas du point general */
	    do { if (++cgl->dir > 5) cgl->dir -= 6; if (debut != adr) *adr = 3;}
	    while ((*(adr = debut + mm0_depla (cgl, image)) & 1));
	    
	} else {

	    /* cas du point de depart */
	    if (cgl->dir > 5) cgl->dir -= 6;
	    while (*(adr = debut + mm0_depla (cgl, image)) & 1) {
		*adr = 3;
		if (cgl->dir == dirdep) {
		    /* fin de l'algorithme */
		    boucle->longueur = cgl->pt1 - ptdep;
		    boucle->origine.x = lx; boucle->origine.y = ly;
		    boucle->liste = ptdep; boucle->type = 1;
		    *debut = 2;
		    return (RESULTAT_OK);
		}
		if (++cgl->dir > 5) cgl->dir -= 6;
	    }
	    
	}
	debut = adr; *cgl->pt1++ = cgl->dir;
	mm0_xynew (cgl); *debut = 2; cgl->dir += 4;
    }
}

/*******************************************************/
/* ajustement des chaines et lacets de cgl->loop2        */
/* resultat dans cgl->loop1                              */
/* im : image contenant le masque de propagation       */
/* val_trace : tous les lacets et chaines sont traces  */
/*       a la valeur val_trace si >0                   */
/* val_fond : seules les portions de chaines qui se    */
/*       trouvaient sur des pixels a cette valeur sont */
/* eliminees.                                          */
/*******************************************************/
void mm0_adjust_fond (image, cgl, val_trace, val_fond)
     int   val_trace, val_fond;
     GLACE *cgl;
     IMAGE *image;
{
    struct lacet *bouc1, *bouc2;
    PIX c, *lecture;
    DIRECTION *ptend, *ptold, *pt1_temp, *pt2_temp;
    
    bouc2 = cgl->loop2; bouc1 = cgl->loop1;
    cgl->pt2 = cgl->chain2; cgl->pt1 = cgl->chain1;
    cgl->imavide = 1;
    
    while (bouc2->liste != 0) {
	ptend = bouc2->liste + bouc2->longueur;
	cgl->xx = bouc2->origine.x; cgl->yy = bouc2->origine.y;
	lecture = image->bitmap + cgl->xx + (cgl->yy)*(image->tx);
	c = (val_trace < 0) ? bouc2->valeur : val_trace;
	cgl->pt2 = bouc2->liste;
	if (bouc2->type == 1) {
	    
	    /* cas d'un lacet */
	    
	    while (cgl->pt2 < ptend) {
		if (*lecture == val_fond) {
		    /* rotation du lacet -> chaine */
		    pt2_temp = bouc2->liste; pt1_temp = cgl->pt1;
		    while (pt2_temp < cgl->pt2) *pt1_temp++ = *pt2_temp++;
		    pt1_temp = bouc2->liste;
		    while (pt2_temp < ptend) *pt1_temp++ = *pt2_temp++;
		    pt2_temp = cgl->pt1;
		    while (pt1_temp < ptend) *pt1_temp++ = *pt2_temp++;
		    bouc2->origine.x = cgl->xx; bouc2->origine.y = cgl->yy;
		    bouc2->type = 0; break;
		}
		cgl->dir = *cgl->pt2++;
		lecture += mm0_depla (cgl,image);
		(void)mm0_xynew (cgl);
	    }
	    if (bouc2->type == 1) {
		/* si lacet -> recopie et trace */
		cgl->xx = bouc1->origine.x = bouc2->origine.x;
		cgl->yy = bouc1->origine.y = bouc2->origine.y;
		bouc1->longueur = bouc2->longueur;
		bouc1->valeur = bouc2->valeur;
		bouc1->type = 1; bouc1->liste = cgl->pt1;
		lecture = image->bitmap + cgl->xx + (cgl->yy)*(image->tx);
		*lecture = c; /* corrige le 24/4/98 pour des points */
		for(cgl->pt2 = bouc2->liste; cgl->pt2 < ptend;) {
		    cgl->dir = *cgl->pt2; lecture += mm0_depla (cgl,image);
		    *lecture = c; (void)mm0_xynew (cgl);
		    *cgl->pt1++ = *cgl->pt2++;
		}
		bouc2++; bouc1++; cgl->imavide = 0;
	    }
	} else {
	    
	    /* cas de la chaine */
	    
	    for (;;) {
		cgl->dir = *cgl->pt2;
		/* recherche de debut de chaine */
		while(cgl->pt2 < 
		      ptend && *(lecture += mm0_depla (cgl,image)) == val_fond) {
		    (void)mm0_xynew (cgl);
		    cgl->pt2++; cgl->dir = *cgl->pt2;
		}
		bouc1->longueur = 0;
		if (cgl->pt2 == ptend) break;
		*lecture = c;
		bouc1->origine.x = cgl->xx; bouc1->origine.y = cgl->yy;
		bouc1->type = 0; bouc1->liste = cgl->pt1;
		ptold = cgl->pt2;
		(void)mm0_xynew (cgl); *cgl->pt1++ = *cgl->pt2++;
		cgl->dir = *cgl->pt2;
		/* suivi de la chaine nouvelle */
		while(cgl->pt2 < ptend && 
		      *(lecture += mm0_depla (cgl,image)) != val_fond) {
		    *cgl->pt1++ = *cgl->pt2; *lecture = c;
		    (void)mm0_xynew (cgl);
		    cgl->pt2++; cgl->dir = *cgl->pt2;
		}
		(void)mm0_xynew (cgl);
		*cgl->pt1++ = *cgl->pt2++;
		bouc1->valeur = bouc2->valeur;
		bouc1->longueur = cgl->pt2 - ptold; bouc1++;
		cgl->imavide = 0;
		if (cgl->pt2 == ptend) break;
	    }
	    bouc2++;
	}
    }
    bouc1->liste = 0;
}

void mm0_copy_chaine (cgl, val1, val2)
     int val1, val2;
     GLACE *cgl;
{
    
    /*********************************************/
    /* copie de boucles(val1) dans boucles(val2) */
    /* et de chaine(val1) dans chaine(val2)      */
    /*********************************************/
    
    int i;
    struct lacet *bouc1, *bouc2;
    
    if (val1 == 1) {
	cgl->pt1 = cgl->chain1; cgl->pt2 = cgl->chain2;
	bouc1 = cgl->loop1; bouc2 = cgl->loop2;
    } else {
	cgl->pt2 = cgl->chain1; cgl->pt1 = cgl->chain2;
	bouc2 = cgl->loop1; bouc1 = cgl->loop2;
    }
    
    while (bouc1->liste != 0) {
	bouc2->origine.x = bouc1->origine.x;
	bouc2->origine.y = bouc1->origine.y;
	bouc2->longueur = bouc1->longueur;
	bouc2->valeur = bouc1->valeur;
        bouc2->liste = cgl->pt2;
	bouc2->type = bouc1->type;
	for (i = 0; i < bouc1->longueur; i++)
	  *cgl->pt2++ = *cgl->pt1++;
	bouc2++; bouc1++;
    }
    bouc2->liste = 0;
}

/******************************************************************/
/* initialisation des chaines correspondant aux points a propager */
/******************************************************************/

void mm0_debut_chaine (cgl, boucle)
     struct lacet *boucle;
     GLACE *cgl;
{

    int xx_temp, yy_temp;

    cgl->dir = *cgl->pt1;
    xx_temp = cgl->xx; yy_temp = cgl->yy;
    (void)mm0_xynew (cgl);
    boucle->origine.x = cgl->xx; boucle->origine.y = cgl->yy;
    boucle->longueur = 2; boucle->type = 0;
    boucle->liste = cgl->pt;
    *cgl->pt++ = (*cgl->pt1 < 3) ? *cgl->pt1 + 3 : *cgl->pt1 - 3;
    *cgl->pt++ = (*cgl->pt2 < 3) ? *cgl->pt2 + 3 : *cgl->pt2 - 3;
    cgl->xx = xx_temp; cgl->yy = yy_temp;
}

/**************************************************************/
/* suivi de tous les contours de l'image im                   */
/* les chaines sont stockees dans cgl->loop1 et cgl->chain1 */
/* im : fond = 0, forme = 1                                   */
/* !! l'image im est modifiee par front.c                     */
/**************************************************************/

int mm0_cont (image, cgl)
     IMAGE *image;
     GLACE *cgl;
{
    
    DIRECTION *ptdep;
    struct lacet *boucle;
    PIX    *pt_im, *pt_im_old;
    int xmin, xmax, ymin, ymax;
    int x, y, ret;
    
    ret = mm0_cadre (image, 0);
    if(ret!=RESULTAT_OK) return(ret);

    xmin = image->lx - 1;
    xmax = xmin + image->sx + 2;
    ymin = image->ly;
    ymax = ymin + image->sy;
    pt_im = image->bitmap + xmin + ymin*image->tx;
    boucle = cgl->loop1;
    ptdep = cgl->chain1;
    
    for (y = ymin; y < ymax; y++) {
	pt_im_old = pt_im;
	for (x = xmin; x < xmax - 1; x++, pt_im++) {
	    if((*pt_im == 0 && (*(pt_im + 1) & 1))
	       || (!(*pt_im & 1) && *(pt_im + 1) == 1)) {
		mm0_front (image, cgl, x + 1, y, 3, boucle, ptdep);
		ptdep += boucle->longueur;
		boucle++;
	    }
	}
	pt_im = pt_im_old + image->tx;
    }
    
    boucle->liste = 0;
    (void) mm0_calgebre (image, 1, image, CG_AND);
    return (RESULTAT_OK);
}

/* cette version trouve les contours exterieurs */

int mm0_cont_e (image, cgl)
     IMAGE *image;
     GLACE *cgl;
{
    DIRECTION *ptdep;
    struct lacet *boucle;
    PIX    *pt_im, *pt_im_old;
    int xmin, xmax, ymin, ymax;
    int x, y, ret;
    
    ret=mm0_cadre (image, 0);
    if(ret!=RESULTAT_OK) return(ret);

    xmin = image->lx - 1;
    xmax = xmin + image->sx + 2;
    ymin = image->ly;
    ymax = ymin + image->sy;
    pt_im = image->bitmap + xmin + ymin*image->tx;

    boucle = cgl->loop1; ptdep = cgl->chain1;
    
    for (y = ymin; y < ymax; y++) {
	pt_im_old = pt_im;
	for (x = xmin; x < xmax - 1; x++, pt_im++) {
	    if((*pt_im == 1 && !(*(pt_im + 1) & 1))
	       || ((*pt_im & 1) && *(pt_im + 1) == 0)) {
		mm0_front_e (image, cgl, x + 1, y, 3, boucle, ptdep);
		ptdep += boucle->longueur;
		boucle++;
	    }
	}
	pt_im = pt_im_old + image->tx;
    }
    
    boucle->liste = 0;
    
    (void)mm0_calgebre (image, 1, image, 5);
    
    return (RESULTAT_OK);
}

void mm0_trace_chaine (image, cgl, val, boucle)
     IMAGE *image;
     int   val;
     GLACE *cgl;
     struct lacet *boucle;
{
    
    int i;
    PIX c, *debut;
    
    c = val;
    
    while (boucle->liste != 0) {
	
	cgl->xx = boucle->origine.x; cgl->yy = boucle->origine.y;
	cgl->pt = boucle->liste;
	debut = image->bitmap + cgl->xx + cgl->yy*image->tx;
	
	if (boucle->type == 1) *debut = c;
	
	for (i = 0; i < boucle->longueur - 1; i++) {
	    cgl->dir = *cgl->pt;
	    debut += mm0_depla (cgl, image);
	    *debut = c; mm0_xynew (cgl);
	    cgl->pt++;
	}
	boucle++;
    }
}
/***********************************************/
/* trace boucle<num> dans l'image <im>         */
/* a la valeur <val> si le pixel correspondant */
/* a la valeur <val_masque> uniquement         */
/***********************************************/

void mm0_trace_masque (image, cgl, val, num, val_masque)
     int   num, val_masque;
     IMAGE *image;
     int   val;
     GLACE *cgl;
{
    
    int i;
    PIX c, c_masque, *debut;
    struct lacet *boucle;
    
    c = val; c_masque = val_masque;
    
    if (num == 1) boucle = cgl->loop1;
    if (num == 2) boucle = cgl->loop2;
    
    while (boucle->liste != 0) {
	
	cgl->xx = boucle->origine.x; cgl->yy = boucle->origine.y;
	cgl->pt = boucle->liste;
	debut = image->bitmap + cgl->xx + cgl->yy*image->tx;
	
	if (boucle->type == 1 && *debut == c_masque) *debut = c;
	
	for (i = 0; i < boucle->longueur - 1; i++) {
	    cgl->dir = *cgl->pt; debut += mm0_depla (cgl, image);
	    if (*debut == c_masque) *debut = c;
	    mm0_xynew (cgl); cgl->pt++;
	}
	boucle++;
    }
}

/*************************************************/
/* regles de recriture pour l'hexagone conjugue  */
/* de taille un demi				 */
/*************************************************/

void mm0_recri_dil_conj ( cgl)
     GLACE *cgl;
{
    int dif;

    dif = (*cgl->pt2 - *cgl->pt1); if (dif < 0) dif += 6;
    switch (dif)
	{
	case 0 : *cgl->pt++ = *cgl->pt1;
	    break;
        case 1 : *cgl->pt++ = *cgl->pt2;
	    *cgl->pt++ = *cgl->pt1;
	    break;
        case 2 : *cgl->pt++ = (*cgl->pt1 < 5) ? *cgl->pt1 + 1 : *cgl->pt1 - 5;
	    *cgl->pt++ = (*cgl->pt1 < 5) ? *cgl->pt1 + 1 : *cgl->pt1 - 5;
	    break;
       case 3 : *cgl->pt++ = (*cgl->pt1 < 5) ? *cgl->pt1 + 1 : *cgl->pt1 - 5;
	    *cgl->pt++ = (*cgl->pt1 < 4) ? *cgl->pt1 + 2 : *cgl->pt1 - 4;
	    break;
       case 4 : *cgl->pt++ = (*cgl->pt1 < 4) ? *cgl->pt1 + 2 : *cgl->pt1 - 4;
	    break;
       case 5 : break;
	}
}

/**************************************************/
/* dilatation par un hexagone conjugue de         */
/* taille un demi de toutes 		          */
/* les chaines de cgl->loop1 dans cgl->loop2    */
/* cgl->loop1 et cgl->chain1 restent intacts    */
/**************************************************/

int mm0_dil_chaine_conj (cgl)
     GLACE *cgl;
{
    struct lacet *boucle1, *boucle2;
    int i;

    boucle1 = cgl->loop1; boucle2 = cgl->loop2; cgl->pt = cgl->chain2;

    while (boucle1->liste != 0) {

	/* cas du point isole */

	if (boucle1->longueur == 0) {
	  /* return (CONJ_SINGLE_POINT); */
	  /* corrige le 24/2/98 */
	  boucle2--;
	} else {
	    boucle2->liste = cgl->pt; boucle2->type = boucle1->type;
	    boucle2->valeur = boucle1->valeur;
	    if (boucle1->type == 1) {

		/* cas du lacet */

		cgl->pt1 = boucle1->liste + boucle1->longueur - 1;
		cgl->pt2 = boucle1->liste;
		cgl->xx = boucle1->origine.x; cgl->yy = boucle1->origine.y;
		cgl->dir = (*cgl->pt1 < 2) ?
		  *cgl->pt1 + 4 : *cgl->pt1 - 2; mm0_xynew (cgl);
		boucle2->origine.x = cgl->xx; boucle2->origine.y = cgl->yy;
		for (i = 0; i < boucle1->longueur; i++) {
		    mm0_recri_dil_conj (cgl);
		    cgl->pt1 = cgl->pt2; cgl->pt2++;
		}
	    } else {

		/* cas general d'une chaine non vide */

		cgl->pt1 = boucle1->liste;
		cgl->pt2 = boucle1->liste + 1;
		boucle2->origine.x = boucle1->origine.x;
		boucle2->origine.y = boucle1->origine.y;
		*cgl->pt++ = (*cgl->pt1 < 1) ? *cgl->pt1 + 5 : *cgl->pt1 - 1;
		for (i =0; i < boucle1->longueur - 1; i++) {
		    mm0_recri_dil_conj (cgl);
		    cgl->pt1 = cgl->pt2; cgl->pt2++;
		}
		*cgl->pt++ = (*cgl->pt1 < 5) ? *cgl->pt1 + 1 : *cgl->pt1 - 5;
	    }
	}
	boucle2->longueur = cgl->pt - boucle2->liste;
	boucle2++; boucle1++;
    }

    boucle2->liste = 0;

    return (RESULTAT_OK);
}

/*******************************/
/* Sens du lacet :             */
/*  6 : autour d'une particule */
/* -6 : autour d'un trou       */
/*******************************/

int mm0_sens (boucle, cgl)
     struct lacet *boucle;
     GLACE  *cgl;
{
    DIRECTION *ptend;
    int deviation, d;

    if (boucle->longueur == 0) return (6);

    ptend = boucle->liste + boucle->longueur;
    d = *boucle->liste - *(ptend - 1); if (d < 0) d += 6;
    deviation = (d > 3) ? d - 6 : d;

    cgl->pt2 = boucle->liste + 1; cgl->pt1 = boucle->liste;
    while (cgl->pt2 < ptend) {
	d = *cgl->pt2++ - *cgl->pt1++; if (d < 0) d += 6;
	deviation += (d > 3) ? d - 6 : d;
    }

    return (deviation);
}

/************************************************************/
/*       TRANSFORME LA FENETRE DE TOUTES LES IMAGES         */
/*       SELON LES CARACTERISTIQUES DE L'IMAGE im           */
/************************************************************/

void set_all_headers (pool, nombre, im)
     IMAGE **pool;
     IMAGE *im;
     int nombre;
{
  int i;

  for (i = 0; i < nombre; i++) {
    if (pool[i] != im) set_header (pool[i], im);
  }
}
/* TRANSFORME UNE SEULE IMAGE */
void set_header (im2, im1)
     IMAGE *im2, *im1;
{
  im2->lx = im1->lx;
  im2->ly = im1->ly;
  im2->sx = im1->sx;
  im2->sy = im1->sy;
}
