#include "morpho.h"

/************************************************************/
/*                                                          */
/*              Lignes de partage des eaux                  */
/*                                                          */
/************************************************************/
/* Watch out: LPE_MAX defines size of local static array */
#define LPE_HMAX   512 /* valeur de gris max ; (h_min = 0) */
#define WSHED   0      /* valeur finale des pixels de la LPE */
#define INIT    -1     /* valeur initiale des pixels de J (imaout) */
#define MASK    -2     /* valeur initiale affectee a chaque niveau */
#define INQUEUE -3     /* valeur d'un pixel a son insertion dans la file */

/* linincr is a copy made on function entry */
#define DIR_INCR_0     1
#define DIR_INCR_1     -linincr + pr
#define DIR_INCR_2     -linincr + pr - 1
#define DIR_INCR_3     -1
#define DIR_INCR_4     linincr + pr - 1
#define DIR_INCR_5     linincr + pr

/* Fonctions de pile */
#define fifo_add(entry, row)         \
    (queue[h_pt] = (entry),          \
     q_row[h_pt++] = (row),          \
     (h_pt>=Q_LEN) ? h_pt-=Q_LEN : 0)

#define fifo_first(entry, row)      \
    (entry = queue[t_pt],           \
     row = q_row[t_pt++],           \
     (t_pt>=Q_LEN) ? t_pt-=Q_LEN : 0)

/* Fonction de traitement d'un point */
#define traite_point(dir, parity)                                \
    if (*(pp = p + (dir)) > 0) {                                 \
	if ((*p == INQUEUE) || ((*p == WSHED) && (flag == 1)))   \
	    *p = *pp;                                            \
	else if ((*p > 0) && (*p != *pp)) {                      \
	    *p = WSHED;                                          \
	    flag = 0;                                            \
	}                                                        \
    } else if (*pp == WSHED) {                                   \
	if (*p == INQUEUE) {                                     \
	    *p = WSHED;                                          \
	    flag = 1;                                            \
	}                                                        \
    } else if (*pp == MASK) {                                    \
	*pp = INQUEUE;                                           \
	fifo_add(pp, parity);                                    \
    }
/************************************************************/
static PIX **queue;
static char *q_row, pr;
static int h_pt, t_pt;
static int Q_LEN;

/* La Ligne de Partage des Eaux premier algorithme */
/* pour utiliser dans une trame hexagonale */

int mm0_lpe1 (image_in, image_out)
     IMAGE *image_in, *image_out;
{
    static int hi[LPE_HMAX + 1], hc[LPE_HMAX + 1];

    PIX *pt_in, *pt_in_old; 
    PIX **it, *p, *pp;
    register PIX *pixel;

    int row, col, label_courant;
    int offset, h, ind, ret, linincr;
    int xmin,xmax,ymin,ymax;
    char flag;
    char *ir;

    if ((ret = BASIC_CHECK (image_in)) != RESULTAT_OK) return (ret);
    if ((ret = BASIC_CHECK (image_out)) != RESULTAT_OK) return (ret);
    
    if ((ret = SET_SIZE (image_in,image_out)) != RESULTAT_OK) return (ret);
    if ((ret = ALLOC_X_CHECK (image_in,image_out)) != RESULTAT_OK) return (ret);

    pt_in = BEG_PIXPTR(image_in);
    linincr = image_in->tx;

    Q_LEN = image_in->sx * image_in->sy;

    queue = (PIX **) malloc (Q_LEN * sizeof (PIX *));
    if (queue == NULL) return(FAILED_MALLOC);
    q_row = (char *) malloc (Q_LEN * sizeof (char));
    if (q_row == NULL) {
	free ((char *) queue);
	return(FAILED_MALLOC);
    }

    it = (PIX **) malloc (Q_LEN * sizeof(PIX *));
    if (it == NULL) {
	free ((char *) queue); free ((char *) q_row);
	return(FAILED_MALLOC);
    }
    ir = (char *)  malloc (Q_LEN * sizeof(char));
    if (ir == NULL) {
	free ((char *) queue); free ((char *) q_row); free ((char *) it);
	return(FAILED_MALLOC);
    }

    mm0_cadre (image_out, INIT);
    xmin = image_in->lx; xmax = xmin + image_in->sx;
    ymin = image_in->ly; ymax = ymin + image_in->sy;
    /*
     * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     * WATCH OUT: access to out done via offset: this assumes that 
     * both images have been allocated contiguously with the same 
     * line size (tx)
     * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     */

    offset = image_out->bitmap - image_in->bitmap;

    /* tri des pixels de i par ordre croissant de niveau de gris */
    for (h = 0; h <= LPE_HMAX; h++) hi[h] = 0;
    hc[0] = 0;

    /* histogramme avec troncature */

    pt_in = image_in->bitmap  + xmin + linincr*ymin;
    for (row = ymin ; row < ymax; row++) {
	pt_in_old = pt_in;
	for (col = xmin ; col < xmax; col++, pt_in++) {
	    if (*pt_in <= 0) hi[0]++;
	    else if (*pt_in >= LPE_HMAX) hi[LPE_HMAX]++;
	    else hi[*pt_in]++;
	    /* Initialisation de Out */
	    *(pt_in + offset) = INIT;
	}
	pt_in = pt_in_old + linincr;
    }
    
    /* histogramme cumule */

    for (h = 1; h <= LPE_HMAX; h++)
	hc[h] = hc[h - 1] + hi[h - 1];

    /* tri proprement dit */

    pt_in = image_in->bitmap + xmin + linincr*ymin;
    for (row = ymin ; row < ymax; row++) {
	pt_in_old = pt_in;

	for (col = xmin ; col < xmax; col++, pt_in++) {
	    it[hc[*pt_in]] = pt_in + offset;
	    ir[hc[*pt_in]] = (row & 1);
	    hc[*pt_in]++;
	}
	pt_in = pt_in_old  + linincr;
    } 

    for (h = LPE_HMAX; h > 0; h--)
	hc[h] = hc[h - 1];
    hc[0] = 0;

    /* LPE proprement dite */
    label_courant = 0;
    flag = 0;
    h_pt = t_pt = 0;

    for (h = 0; h <= LPE_HMAX; h++) 
    {
	/* SKIZ geodesique di niveau h-1 dans le niveau h */
	for (ind = hc[h]; ind < hc[h + 1]; ind++) {
	    pp  = it[ind];
	    pr = ir[ind];

	    /* ces pixels sont accessibles directement grace au trial initial */
	    *pp = MASK;
	    if (   (*(pp + DIR_INCR_0) > 0) || (*(pp + DIR_INCR_0) == WSHED) 
		|| (*(pp + DIR_INCR_1) > 0) || (*(pp + DIR_INCR_1) == WSHED) 
		|| (*(pp + DIR_INCR_2) > 0) || (*(pp + DIR_INCR_2) == WSHED) 
		|| (*(pp + DIR_INCR_3) > 0) || (*(pp + DIR_INCR_3) == WSHED) 
		|| (*(pp + DIR_INCR_4) > 0) || (*(pp + DIR_INCR_4) == WSHED) 
		|| (*(pp + DIR_INCR_5) > 0) || (*(pp + DIR_INCR_5) == WSHED))
	    {
		*pp = INQUEUE;
		fifo_add(pp, pr);
	    }
	}

	while (h_pt != t_pt) 
	{
	    fifo_first(p, pr);

	    traite_point(DIR_INCR_0, pr);
	    traite_point(DIR_INCR_1, 1 - pr);
	    traite_point(DIR_INCR_2, 1 - pr);
	    traite_point(DIR_INCR_3, pr);
	    traite_point(DIR_INCR_4, 1 - pr);
	    traite_point(DIR_INCR_5, 1 - pr);

	}

	/* de nouveaux minima sont ils apparus ? */
	for (ind = hc[h]; ind < hc[h + 1]; ind++) {
	    pixel  = it[ind];
	    pr = ir[ind];
	    if (*pixel == MASK) {
		label_courant++;
		fifo_add(pixel, pr);
		*pixel = label_courant;
		while (h_pt != t_pt) {
		    fifo_first(p, pr);
		    if (*(pp = p + DIR_INCR_0) == MASK) {
			fifo_add(pp, pr);
			*pp = label_courant;
		    }
		    if (*(pp = p + DIR_INCR_1) == MASK) {
			fifo_add(pp, 1 - pr);
			*pp = label_courant;
		    }
		    if (*(pp = p + DIR_INCR_2) == MASK) {
			fifo_add(pp, 1 - pr);
			*pp = label_courant;
		    }
		    if (*(pp = p + DIR_INCR_3) == MASK) {
			fifo_add(pp, pr);
			*pp = label_courant;
		    }
		    if (*(pp = p + DIR_INCR_4) == MASK) {
			fifo_add(pp, 1 - pr);
			*pp = label_courant;
		    }
		    if (*(pp = p + DIR_INCR_5) == MASK) {
			fifo_add(pp, 1 - pr);
			*pp = label_courant;
		    }
		} /* end while (h_ptr != t_ptr) */
	    } /* end if (MASK) */
	}  /* for (ind = hc[h]; ind < hc[h + 1]; ind++) */
    }/* for (h = 1; h <= LPE_HMAX; h++) */

    free ((char *) queue);
    free ((char *) q_row);
    free ((char *) it);
    free ((char *) ir);

    return (RESULTAT_OK);
}
