#include "morpho.h"

/************************************************************/
/*      LA SURFACE...                                       */
/************************************************************/

int mm0_area (im, surface)
     IMAGE *im; int *surface;
{
  int ret, i, j;
  PIX *pt_im, *pt_im_old;

  if ((ret = BASIC_CHECK (im)) != RESULTAT_OK) return (ret);

  *surface = 0;
  pt_im = BEG_PIXPTR(im);
  for (j = 0; j < im->sy; j++) {
    pt_im_old = pt_im;
    for (i = 0 ; i < im->sx; i++, pt_im++)
      if (*pt_im > 0) (*surface)++;
    pt_im = pt_im_old + im->tx;
  }

  return (RESULTAT_OK);
}

/**************************/
/* Miveau de gris minimum */
/**************************/
int mm0_minim (in, minval)
     IMAGE *in;
     int   *minval;
{
    
    register PIX *pt_in;
    PIX *pt_in_old;
    register int i;
    int j, ret;
    register PIX min;

    if ((ret = BASIC_CHECK (in)) != RESULTAT_OK) return (ret);
    
    min = PIX_MAX;
    pt_in = BEG_PIXPTR(in);
    
    for (j = 0; j < in->sy; j++) {
	pt_in_old = pt_in;
	for (i = 0 ; i < in->sx; i++, pt_in++)
	  if (*pt_in < min) min = *pt_in;

	pt_in = pt_in_old + in->tx;
    }
    *minval=(int)min;
    return (RESULTAT_OK);
}

/**************************/
/* Miveau de gris maximum */
/**************************/
int mm0_maxim (in, maxval)
     IMAGE *in;
     int   *maxval;
{
    
    register PIX *pt_in;
    PIX *pt_in_old;
    register int i;
    int j, ret;
    register PIX max;

    if ((ret = BASIC_CHECK (in)) != RESULTAT_OK) return (ret);
    
    max = PIX_MIN;
    pt_in = BEG_PIXPTR(in);
    
    for (j = 0; j < in->sy; j++) {
	pt_in_old = pt_in;
	for (i = 0 ; i < in->sx; i++, pt_in++)
	  if (*pt_in > max) max = *pt_in;

	pt_in = pt_in_old + in->tx;
    }
    *maxval=(int)max;
    return (RESULTAT_OK);
}

/*******************/
/* Nombre d'Euler  */
/*******************/
int mm0_euler (im_in, val)
     IMAGE *im_in;
     int  *val;
{
    GLACE  *cgl, *mm0_alloc_chains();
    struct lacet *boucle;
    int nb = 0, ret;

    /* loops & chains allocated in label, use fixed values */
    cgl=mm0_alloc_chains (im_in);
    if(cgl==NULL) return(LACET_ALLOC_FAILED);

    ret=mm0_cont (im_in, cgl);
    if(ret!=RESULTAT_OK) return(ret);

    boucle = cgl->loop1;
    while (boucle->liste != 0) {
	if (mm0_sens(boucle,cgl) == 6) nb++;
	else nb--;
	boucle++;
    }
    *val = nb;
    return (RESULTAT_OK);
}

/*******************/
/* Nombre d'objets */
/*******************/
int mm0_count_objects (im_in, val)
     IMAGE *im_in;
     int *val;
{
    GLACE  *cgl, *mm0_alloc_chains();
    struct lacet *boucle;
    int nb = 0, ret;

    /* loops & chains allocated in label, use fixed values */
    cgl=mm0_alloc_chains (im_in);
    if(cgl==NULL) return(LACET_ALLOC_FAILED);
    
    ret=mm0_cont (im_in, cgl);
    if(ret!=RESULTAT_OK) return(ret);
    
    boucle = cgl->loop1;
    while (boucle->liste != 0) {
	if (mm0_sens(boucle,cgl) == 6) nb++;
	boucle++;
    }
    *val = nb;
    return (RESULTAT_OK);
}

/*************/
/* Perimetre */
/*************/
int mm0_perimetre (im_in, val)
     IMAGE *im_in;
     int *val;
{
    GLACE  *cgl, *mm0_alloc_chains();
    struct lacet *boucle;
    int nb = 0, ret;

    /* loops & chains allocated in label, use fixed values */
    cgl=mm0_alloc_chains (im_in);
    if(cgl==NULL) return(LACET_ALLOC_FAILED);

    ret=mm0_cont (im_in, cgl);
    if(ret!=RESULTAT_OK) return(ret);
    
    boucle = cgl->loop1;
    while (boucle->liste != 0) {
	nb += boucle->longueur;
	boucle++;
    }
    *val = nb;
    return (RESULTAT_OK);
}

    /********************************************/
    /* Nombre de pixels verifiant une condition */
    /* type = 1 : image (in) <  cte     */
    /*        2 : image (in) <= cte     */
    /*        3 : image (in) == cte     */
    /*        4 : image (in) >= cte     */
    /*        5 : image (in) >  cte     */
    /*        6 : image (in) != cte     */
    /************************************/
int mm0_compter (in, cte, res_count, type)
     IMAGE *in;
     int   cte;
     long  *res_count;
     int   type;
{
    PIX *pt_in, *pt_in_old;
    int i, j, ret;
    register long count;
    
    if ((ret = BASIC_CHECK (in)) != RESULTAT_OK) return (ret);
    
    pt_in = BEG_PIXPTR(in);
    count = 0;
    
    for (j = 0; j < in->sy; j++) {
	pt_in_old = pt_in;
	switch (type)
	  {
	      default:
	      return (PARAM_NOT_IN_RANGE);
	      
	      case CG_LT :
		for (i = 0 ; i < in->sx; i++)
		  if (*pt_in++ < cte) count++;
	      break;
	      case CG_LE : 
		for (i = 0 ; i < in->sx; i++)
		  if (*pt_in++ <= cte) count++;
	      break;
	      case CG_EQ : 
		for (i = 0 ; i < in->sx; i++)
		  if (*pt_in++ == cte) count++;
	      break;
	      case CG_NEQ : 
		for (i = 0 ; i < in->sx; i++)
		  if (*pt_in++ >= cte) count++;
	      break;
	      case CG_GE : 
		for (i = 0 ; i < in->sx; i++)
		  if (*pt_in++ > cte) count++;
	      break;
	      case CG_GT : 
		for (i = 0 ; i < in
->sx; i++)
		  if (*pt_in++ != cte) count++;
	      break;
	  }
	pt_in = pt_in_old + in->tx;
    }
    *res_count = count;
    return (RESULTAT_OK);
}

int mm0_histo (image_in, min, max, mat)
     IMAGE *image_in;
     int min, max, *mat;
{
/* histogramme de l'image im */
/* les valeurs < min sont comtees avec min */
/* de meme pour max */
/* resultat dans mat */
    int xmin, xmax, ymin, ymax;
    register PIX *pt_in;
    PIX *pt_in_old;
    int i, j, ret;

    if ((ret = BASIC_CHECK (image_in)) != RESULTAT_OK) return (ret);

    pt_in = image_in->bitmap + image_in->lx + image_in->tx*image_in->ly;
    xmin = image_in->lx; xmax = xmin + image_in->sx;
    ymin = image_in->ly; ymax = ymin + image_in->sy;

    /* histogramme baseline */
    for (i = 0; i <= max - min; i++) mat[i] = 0;

    for (j = ymin; j < ymax; j++) {
	pt_in_old = pt_in;
	for (i = xmin ; i < xmax; i++, pt_in++) {
	    if (*pt_in < min) mat[0]++;
	    else if (*pt_in > max) mat[max-min]++;
	    else mat[(*pt_in) - min]++;
	}
	pt_in = pt_in_old + image_in->tx;
    }

    return (RESULTAT_OK);
}
