/* beg_algo_entry

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% ipl_skiz
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\begin{algorithme}
{\bf Algorithme~: Squelette par zones d'influence (Skiz)}
\medskip
\begin{verbatim}
int ipl_skiz (ipl_image *pt_in, ipl_image **pt_out);
\end{verbatim}
\separ
Entr\'ee~:
	\begin{itemize}
	\item[] $I$, image \'etiquet\'ee;
	\end{itemize}
	\centerline{/ * Le skiz est calcul\'e directement dans $I$ * /}
Sortie~:
	\begin{itemize}
	\item[] $I$, image \'etiquet\'ee;
	\end{itemize} 
\separ
Suivre le contour de chaque composante connexe de $I$ en
conservant sa valeur\\
Tant qu'il reste des cha\^{\i}nes ou des lacets \{\\
	\begin{itemize}
	\item[] Dilater les cha\^{\i}nes et lacets
	\item[] Mettre \`a jour, $Fond=0$, $Forme=$valeur de la cha\^{\i}ne
	\end{itemize} 
\}\\
\end{algorithme}
   end_algo_entry */
   
#include "morpho.h"
   
   /************************************************/
   /* Zones d'influence des particules             */
   /* le resultat est une image labellee des zones */
   /************************************************/
   
int mm0_skiz (inout, out)
     IMAGE *inout, *out;
 {
     DIRECTION *ptdep;
     struct lacet *boucle;
     PIX    *pt_im, *pt_im_old;
     GLACE  *cgl, *mm0_alloc_chains();
     int    x, y, xmin, xmax, ymin, ymax, ret;
     
     /* all basic checks done by copy */
     if((ret = mm0_copy (inout, out))!=RESULTAT_OK) return(ret);
     (void)mm0_cadre (inout, 0);
     (void)mm0_cadre (out, 0);
     if((ret = mm0_label (out))!=RESULTAT_OK) return(ret);

     xmin = inout->lx; xmax = xmin + inout->sx;
     ymin = inout->ly; ymax = ymin + inout->sy;

     pt_im = inout->bitmap + xmin + ymin*inout->tx;
     /* loops & chains allocated in label, use fixed values */
     cgl=mm0_alloc_chains (inout); /* can't fail, done in label too */
     boucle = cgl->loop1;
     ptdep = cgl->chain1;
     
     for (y = ymin; y < ymax; y++) {
	 pt_im_old = pt_im;
	 for (x = xmin; x < xmax - 1; x++, pt_im++) {
	     int pix_val;
	     if((*pt_im == 0 && (*(pt_im + 1) & 1))
		|| (!(*pt_im & 1) && *(pt_im + 1) == 1)) {
		 (void)mm0_front (inout, cgl, x + 1, y, 3, boucle, ptdep);
		 if((ret = mm0_lire (out, x+1, y, &pix_val))!=RESULTAT_OK)
		   return(ret);
		 boucle->valeur = (int)pix_val;
		 /* printf ("[val:%d]", boucle->valeur); fflush (stdout); */
		 ptdep += boucle->longueur;
		 boucle++;
	     }
	 }
	 pt_im = pt_im_old + inout->tx;
     }

     boucle->liste = 0;
     mm0_calgebre (inout, (PIX)1, inout, CG_AND); /*  pix[i] &= 1; */
     mm0_cadre (out, 1);
     cgl->imavide = 0;
     
     while (cgl->imavide == 0) {
	 mm0_dil_chaine (cgl);
	 mm0_adjust_masque (out, cgl, -1, 0);
	 /* printf ("*"); fflush (stdout); */
     }
     return (RESULTAT_OK);
 }
