#include "morpho.h"
static int Q_LEN_SQU1;

/* squelettisation : premier algorithme :  pour une trame hexagonale */

int mm0_squ1 (image_inout, ancrage)
     IMAGE *image_inout, *ancrage;
{
    static char tab_homo [64] ={
	1, 0, 0, 0, 0, 1, 0, 0,
	0, 1, 1, 1, 0, 1, 0, 0,
	0, 1, 1, 1, 1, 1, 1, 1,
	0, 1, 1, 1, 0, 1, 0, 0,
	0, 0, 1, 0, 1, 1, 1, 0, 
	1, 1, 1, 1, 1, 1, 1, 0,
	0, 0, 1, 0, 1, 1, 1, 0,
	0, 0, 1, 0, 0, 0, 0, 1 };

static PIX **queue;
static char *q_row, pr;

    PIX *pt_in, *pt_in_old, *pt_anc, *pt_anc_old; 
    PIX *p, *pp, *pix_fictif;

    int i, row, col, h_ptr, t_ptr;
    int ret, lincr, xmin, xmax, ymin, ymax;
    char flag, stabilite, index;
    
    if ((ret = BASIC_CHECK (image_inout)) != RESULTAT_OK) return (ret);
    if ((ret = BASIC_CHECK (ancrage)) != RESULTAT_OK) return (ret);    
    if ((ret = ALLOC_X_CHECK (image_inout, ancrage))
	!= RESULTAT_OK) return (ret);
    if ((ret = SET_SIZE (image_inout, ancrage)) != RESULTAT_OK) return (ret);

    Q_LEN_SQU1 = image_inout->sx * image_inout->sy;

    queue = (PIX **) malloc (Q_LEN_SQU1 * sizeof (PIX *));
    if (queue == NULL) return(FAILED_MALLOC);
    q_row = (char *) malloc (Q_LEN_SQU1 * sizeof (char));
    if (q_row == NULL) {
	free ((char *) queue);
	return(FAILED_MALLOC);
    }

    (void)mm0_cadre (ancrage, PIX_MAX);

    xmin = image_inout->lx; xmax = xmin + image_inout->sx;
    ymin = image_inout->ly; ymax = ymin + image_inout->sy;

    /* squelette */
    stabilite = 1;
    lincr = image_inout->tx;

    /*
     * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     * WATCH OUT: access to out done via lincr: this assumes that 
     * both images have been allocated contiguously with the same 
     * line size (tx)
     * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     */

    pt_in = image_inout->bitmap  + xmin + lincr*ymin;
    pt_anc= ancrage->bitmap + xmin + lincr*ymin;
    
    /* pour tout pixel */
    for (row = ymin ; row < ymax; row++) {
	pt_in_old  = pt_in;
	pt_anc_old = pt_anc;
	for (col = xmin ; col < xmax; col++, pt_in++, pt_anc++) {
	    
	    /* cas des points d'ancrage */
	    if (*pt_anc == 1) 
	      *pt_in = 1;
	    
	    /* points qui ne sont pas points d'ancrage */
	    else
	      if (*pt_in == 1)
		*pt_in = 2;
	    
	} /* end for (col) */
	
	pt_in  = pt_in_old  + lincr;
	pt_anc = pt_anc_old + lincr;
	
    } /* end for (row) */
    
    /* initialisation de la file d'attente */
    h_ptr = 0;
    t_ptr = 0;
    
    pt_in = image_inout->bitmap  + xmin + lincr*ymin;
    
    for (row = ymin ; row < ymax; row++) {
	pt_in_old  = pt_in;
	for (col = xmin ; col < xmax; col++, pt_in++) {
	    if (*pt_in == 2){
		
		/* verification des points de voisinage de fond */
		flag = 0;
		if (*(pt_in + 1)                    == 0) flag = 1;
		if (*(pt_in - lincr + (row & 1))     == 0) flag = 1;
		if (*(pt_in - lincr + (row & 1) - 1) == 0) flag = 1;
		if (*(pt_in - 1)                    == 0) flag = 1;
		if (*(pt_in + lincr + (row & 1) - 1) == 0) flag = 1;
		if (*(pt_in + lincr + (row & 1))     == 0) flag = 1;
		
		if (flag == 1){
		    *pt_in = 3;
		    fifo_add_1(queue, q_row, &h_ptr, pt_in, (row & 1));  	  
		    
		} /* end if (=1) */
		
	    }/* end if (=2) */
	    
	} /* end for (col) */
	
	pt_in  = pt_in_old  + lincr;
	
    } /* end for (row) */
    
    /* add pix_fictif */
    fifo_add_1(queue, q_row, &h_ptr, pix_fictif, 0);
    
    /* propagation */
    
    for (;;){
	
	fifo_first_1(queue, q_row, &t_ptr, &p, &pr);
	
	if (p == pix_fictif){
	    
	    /* queue empty if t_ptr == h_ptr */
	    if (t_ptr != h_ptr){
		if (stabilite == 1){
		    
		    /* mise a 1 des points du squelette */
		    while (t_ptr != h_ptr){
			fifo_first_1(queue, q_row, &t_ptr, &pp, &pr);
	      		*pp = 1;
			
		    } /* end while */
		    
		    /* sortie brutale de la boucle infinie */
		    break;
		    
		} /* end if (stabilite) */
		
		else{
		    stabilite = 1;
		    fifo_add_1(queue, q_row, &h_ptr, pix_fictif, 0);
		    
		} /* end else */
	    }
	    else
	      /* la file d'attente est vide */
	      break;
	    
	} /* end if (pix_fictif) */
	
	else{
	    /* le pixel p n'est pas fictif */
	    index = 0;
	    
	    /* calcul du code de la configuration de voisinage de p */
	    /* pr is memory of (row & 1) */
	    
	    for (i = 1; i < 64; i += i ){
		switch (i){
		    case 1:
		    pp = p + 1;
		    break;
		    case 2:
		    pp = p - lincr + pr;
		    break;
		    case 4:
		    pp = p - lincr + pr - 1;
		    break;
		    case 8:
		    pp = p - 1;
		    break;
		    case 16:
		    pp = p + lincr + pr - 1;
		    break;
		    case 32:
		    pp = p + lincr + pr;
		    break;
		} /* end switch */
		
		if (*pp > 0)  index += i;
		
		if (*pp == 2) {
		    if ((i == 1) || (i == 8))
		      fifo_add_1(queue, q_row, &h_ptr, pp, pr);
		    else
		      fifo_add_1(queue,
				 q_row, &h_ptr, pp, (pr == 0) ? 1 : 0);
		    *pp = 3;
		}
		
	    } /* end for (i) */
	    
	    /* on remet sur la file les pixels "non eliminables" */
	    if (tab_homo[index] == 1)
	      fifo_add_1(queue, q_row, &h_ptr, p, pr);
	    
	    else{
		*p = 0;
		stabilite = 0;
		
	    } /* end else */
	    
	} /* end else (pas fictif) */
	
    } /* end for (;;) */

    /* it is possible that the anchor points have been defined with loops
       so that some of the pixels will not be considered and the
       algorithm will assign them at value 2,
       hence a clean up of the result is indicated */

   /* level 2 mapped onto 1 */
    (void)mm0_thresh(image_inout, image_inout, 1, 3);
    free ((char *) queue);
    free ((char *) q_row);
    return (RESULTAT_OK);
}

/* =========================================================== */

static fifo_add_1 (q_main, q_row, h_pt, entry, row)
     PIX  *q_main[], *entry;
     char q_row[], row;
     int  *h_pt;
{
    /* adds an element to the fifo Q */
    q_main[*h_pt] = entry;
    q_row[*h_pt]  = row;
    *h_pt = (*h_pt + 1) % Q_LEN_SQU1;
}

/* =========================================================== */

static fifo_first_1 (q_main, q_row, t_pt, entry, row)
     PIX *q_main[], **entry;
     char q_row[], *row;
     int *t_pt;
{
    /* returns the first element from the queue */
    *entry = q_main[*t_pt];
    *row = q_row[*t_pt];
    *t_pt = (*t_pt + 1) % Q_LEN_SQU1;
}

/************************************************************/
/*                                                          */
/*   Squelette version 2 (rotation de la file d'attente     */
/*                                                          */
/************************************************************/

#define fifo_add_2(q_main,q_row,h_pt,dir,entry,row) \
  (q_main)[(h_pt)]=(entry), \
  (q_row)[(h_pt)]=(row), \
  (h_pt)=(((h_pt)+(dir))==-1)?(qlen):((h_pt)+(dir))%(qlen)

#define fifo_first_2(q_main,q_row,t_pt,dir,entry,row) \
  (entry)=(q_main)[(t_pt)], \
  (row)=(q_row)[(t_pt)], \
  (t_pt)=(((t_pt)+(dir))==-1)?(qlen):((t_pt)+(dir))%qlen

#define fifo_invert_2(h_pt,t_pt,dir) \
  { int temp; \
	(dir)  = ((dir) == 1) ? -1 : 1; \
	(temp)  = (t_pt); \
    (t_pt) = (h_pt) + (dir); \
    (h_pt) = (temp) + (dir); \
  }

#define VOISIN0(bp,br) pp=(bp)+1;jr=(br)
#define VOISIN1(bp,br) pp=(bp)+(br)+im_in->tx;jr=1-(br)
#define VOISIN2(bp,br) pp=(bp)+(br)+im_in->tx-1;jr=1-(br)
#define VOISIN3(bp,br) pp=(bp)-1;jr=(br)
#define VOISIN4(bp,br) pp=(bp)+(br)-im_in->tx-1;jr=1-(br)
#define VOISIN5(bp,br) pp=(bp)+(br)-im_in->tx;jr=1-(br)

/* squelettisation : deuxieme algorithme : ref. chpt 7 p.168 */
/* pour une trame hexagonale */

int mm0_squ2 (im_in, im_anc, mode)
     int  mode;
     IMAGE *im_in, *im_anc;
{
    static char tab_homo [64] ={
                            1, 0, 0, 0, 0, 1, 0, 0,
			    0, 1, 1, 1, 0, 1, 0, 0,
			    0, 1, 1, 1, 1, 1, 1, 1,
			    0, 1, 1, 1, 0, 1, 0, 0,
			    0, 0, 1, 0, 1, 1, 1, 0, 
			    1, 1, 1, 1, 1, 1, 1, 0,
			    0, 0, 1, 0, 1, 1, 1, 0,
			    0, 0, 1, 0, 0, 0, 0, 1 };

    static char tab_ebarb [64] ={
	                    0, 0, 0, 0, 0, 1, 0, 1,
			    0, 1, 1, 1, 0, 1, 1, 1,
			    0, 1, 1, 1, 1, 1, 1, 1,
			    0, 1, 1, 1, 1, 1, 1, 1,
			    0, 0, 1, 1, 1, 1, 1, 1,
			    1, 1, 1, 1, 1, 1, 1, 1,
			    0, 1, 1, 1, 1, 1, 1, 1,
			    1, 1, 1, 1, 1, 1, 1, 1 };

    PIX *pt_in, *pt_in_old, *pt_anc, *pt_anc_old; 
    PIX *p, *pp, *pix_fictif, *fict_pix;
    int i, row, col, h_ptr, t_ptr, tab_ptr;
    char flag, index, dir, jr;
    int xmin, xmax, ymin, ymax, ret;

    int qlen, tab_size;

    PIX **tab, **queue;
    char *t_row, *q_row, pr;

    if ((ret = BASIC_CHECK (im_in)) != RESULTAT_OK) return (ret);
    if ((ret = BASIC_CHECK (im_anc)) != RESULTAT_OK) return (ret);
    
    if ((ret = ALLOC_X_CHECK (im_in, im_anc))
	!= RESULTAT_OK) return (ret);
    if ((ret = SET_SIZE (im_in, im_anc)) != RESULTAT_OK) return (ret);

    xmin = im_in->lx; xmax = xmin + im_in->sx;
    ymin = im_in->ly; ymax = ymin + im_in->sy;

	qlen = im_in->sx*im_in->sy;
	tab_size = qlen;
	
    if ((queue = (PIX **)malloc(qlen*sizeof(PIX*))) == NULL)
	  {
	  f1:
		return(FAILED_MALLOC);
	  }
    if ((q_row = (char *)malloc(qlen*sizeof(char))) == NULL)
	  {
	  f2:
		free(queue);
		goto f1;
	  }
    if ((tab = (PIX **)malloc(tab_size*sizeof(PIX*))) == NULL)
	  {
	  f3:
		free(q_row);
		goto f2;
	  }
    if ((t_row = (char *)malloc(tab_size*sizeof(char))) == NULL)
	  {
		free(tab);
		goto f3;
	  }
	
    (void)mm0_cadre (im_anc, (mode == 0) ? 255 : 0); 

    pt_in = im_in->bitmap  + xmin + im_in->tx*ymin;
    pt_anc= im_anc->bitmap + xmin + im_in->tx*ymin;

    /* pour tout pixel */
    for (row = ymin ; row < ymax; row++) {
	pt_in_old  = pt_in;
	pt_anc_old = pt_anc;
	for (col = xmin ; col < xmax; col++, pt_in++, pt_anc++) {

	    /* cas des points d'ancrage */
	    if (*pt_anc == 1) 
		*pt_in = 1;

	    /* points qui ne sont pas points d'ancrage */
	    else
		if (*pt_in == 1)
		    *pt_in = 2;
	} /* end for (col) */
	pt_in  = pt_in_old  + im_in->tx;
	pt_anc = pt_anc_old + im_in->tx;
    } /* end for (row) */

    /* initialisation de la file d'attente */
    tab_ptr = 0;
    h_ptr   = 0;
    t_ptr   = 0;
    dir     = 1;

    pt_in = im_in->bitmap  + xmin + im_in->tx*ymin;
    for (row = ymin ; row < ymax; row++) {
	  pt_in_old  = pt_in;
	  for (col = xmin ; col < xmax; col++, pt_in++) {
	    if (*pt_in == 2){

		  /* verification des points de voisinage de fond */
		  
		  flag = (*(pt_in + 1) == 0)
			| (*(pt_in - im_in->tx + (row & 1)) == 0)
			  | (*(pt_in - im_in->tx + (row & 1) - 1) == 0)
				| (*(pt_in - 1) == 0)
				  | (*(pt_in + im_in->tx + (row & 1) - 1) == 0)
					| (*(pt_in + im_in->tx + (row & 1)) == 0);

		  if (flag == 1){
		    *pt_in = 3;
		    fifo_add_2(queue, q_row, h_ptr, dir, pt_in, (row & 1));
		  } /* end if (=1) */
	    }/* end if (=2) */
	  } /* end for (col) */
	  pt_in  = pt_in_old  + im_in->tx;
    } /* end for (row) */

    /* add pix_fictif */
    fifo_add_2(queue, q_row, h_ptr, dir, pix_fictif, 0);
     
    /* propagation */
    
    for (;;){
	fifo_first_2(queue, q_row, t_ptr, dir, p, pr);	
	if (p == pix_fictif){
	    
	    /* queue empty if t_ptr == h_ptr */
	    if (t_ptr == h_ptr) {

		/* elimination des barbules parasites restantes */
		/* 1. reperage des extremites de barbules */
		for(i = 0; i < tab_ptr; i++){
		    p  = tab[i];
		    pr = t_row[i];
		    index = vois_cod(p, pr, im_in->tx);
		    if (tab_ebarb[index] == 0){
			*p = 2;
			fifo_add_2(queue, q_row, h_ptr, dir, p, pr);
		    } /* end if */
		} /* end for (i) */

		/* 2. elimination de ces barbules */
		while (h_ptr != t_ptr){
		    fifo_first_2(queue, q_row, t_ptr, dir, p, pr);
		    index = vois_cod(p, pr, im_in->tx);
		    if (tab_ebarb[index] == 0){

			/* elimination de p */
			*p = 0;
#define FAIT(x,y)  if (*pp == (x)) \
			          fifo_add_2(queue, q_row, h_ptr, dir, pp, jr), *pp=(y)
			VOISIN0(p, pr);
			FAIT(3,2);
			VOISIN1(p, pr);
			FAIT(3,2);
			VOISIN2(p, pr);
			FAIT(3,2);
			VOISIN3(p, pr);
			FAIT(3,2);
			VOISIN4(p, pr);
			FAIT(3,2);
			VOISIN5(p, pr);
			FAIT(3,2);
		    } /* end if (tab_ebarb) */
		    else
			*p = 3;
		} /* end while */

		/* 3. les pixels de TAB a la valeur 3 
		   appartiennent au squelette */

		for(i = 0; i < tab_ptr; i++)
		    if (*tab[i] == 3) *tab[i] = 1;
		/* sortie de la boucle infinie */
		break;
	    }
	    else{
		/* la file n'est pas vide */
		fifo_invert_2(h_ptr, t_ptr, dir);
		fifo_add_2(queue, q_row, h_ptr, dir, pix_fictif, 0);
	    } /* end else */
	}
	else{
	    /* le pixel p n'est pas fictif */
	    /* calcul du code de la configuration de voisinage de p */
	    /* pr is memory of (row & 1) */
	    
	  VOISIN0(p, pr);
	  FAIT(2,3);
	  VOISIN1(p, pr);
	  FAIT(2,3);
	  VOISIN2(p, pr);
	  FAIT(2,3);
	  VOISIN3(p, pr);
	  FAIT(2,3);
	  VOISIN4(p, pr);
	  FAIT(2,3);
	  VOISIN5(p, pr);
	  FAIT(2,3);

	  index = vois_cod(p, pr, im_in->tx);

	    /* on met p dans TAB s'il est non eliminable */
	  if (tab_homo[index] == 1){
		tab[tab_ptr]   = p;
		t_row[tab_ptr] = pr;
		tab_ptr++;
	  }
	  else
		*p = 0;
	}
  } /* end for (;;) */

    /* it is possible that the anchor points have been defined with loops
       so that some of the pixels will not be considered and the
       algorithm will assign them at value 2,
       hence a clean up of the result is indicated */

    (void)mm0_thresh(im_in, im_in, 1, 3); /* level 2 mapped onto 1 */

	free(t_row);
	free(tab);
	free(q_row);
	free(queue);

    return (RESULTAT_OK);
}

vois_cod(p, pr, incr)
PIX *p;
int pr, incr;
{
    /* calcul du code de la configuration de voisinage de p */
    /* pr is memory of (row & 1) */
    char index, i;
    PIX *pp;

    index = 0;
    for (i = 1; i < 64; i += i ){
	switch (i){
	case 1:
	    pp = p + 1;
	    break;
	case 2:
	    pp = p - incr + pr;
	    break;
	case 4:
	    pp = p - incr + pr - 1;
	    break;
	case 8:
	    pp = p - 1;
	    break;
	case 16:
	    pp = p + incr + pr - 1;
	    break;
	case 32:
	    pp = p + incr + pr;
	    break;
	} /* end switch */
	
	if (*pp > 0)  index += i;
	
    } /* end for (i) */
    
    return(index);
}    

#undef fifo_add_2
#undef fifo_first_2
#undef fifo_invert_2
#undef VOISIN0
#undef VOISIN1
#undef VOISIN2
#undef VOISIN3
#undef VOISIN4
#undef VOISIN5
#undef FAIT
